r"""
"""
import numpy as _np


def sphere(
    target,
    pore_diameter='pore.diameter',
    throat_cross_sectional_area='throat.cross_sectional_area'
):
    r"""
    Calculates internal surface area of pore bodies assuming they are
    spherical then subtracts the area of the neighboring throats in a
    crude way, by simply considering the throat cross-sectional area, thus
    not accounting for the actual curvature of the intersection.

    Parameters
    ----------
    target : GenericGeometry
        The Geometry object for which these values are being calculated.
        This controls the length of the calculated array, and also
        provides access to other necessary thermofluid properties.
    pore_diameter : str
        The dictionary key to the pore diameter array.
    throat_cross_sectional_area : str
        The dictionary key to the throat cross sectional area array.
        Throat areas are needed since their insection with the pore are
        removed from the computation.

    Returns
    -------
    value : NumPy ndarray
        Array containing pore surface area values.

    """
    network = target.project.network
    R = target[pore_diameter] / 2
    Asurf = 4 * _np.pi * R**2
    Tn = network.find_neighbor_throats(pores=target.Ps, flatten=False)
    Tsurf = _np.array([network[throat_cross_sectional_area][Ts].sum() for Ts in Tn])
    value = Asurf - Tsurf
    return value


def circle(
    target,
    pore_diameter='pore.diameter',
    throat_cross_sectional_area='throat.cross_sectional_area'
):
    r"""
    Calculates internal surface area of pore bodies assuming they are
    circular then subtracts the area of the neighboring throats in a
    crude way, by simply considering the throat cross-sectional area, thus
    not accounting for the actual curvature of the intersection.

    Parameters
    ----------
    target : GenericGeometry
        The Geometry object for which these values are being calculated.
        This controls the length of the calculated array, and also
        provides access to other necessary thermofluid properties.
    pore_diameter : str
        The dictionary key to the pore diameter array.
    throat_cross_sectional_area : str
        The dictionary key to the throat cross sectional area array.
        Throat areas are needed since their insection with the pore are
        removed from the computation.

    Returns
    -------
    value : NumPy ndarray
        Array containing pore surface area values.

    """
    network = target.project.network
    R = target[pore_diameter] / 2
    Asurf = 2 * _np.pi * R
    Tn = network.find_neighbor_throats(pores=target.Ps, flatten=False)
    Tsurf = _np.array([network[throat_cross_sectional_area][Ts].sum() for Ts in Tn])
    value = Asurf - Tsurf
    return value


def cube(
    target,
    pore_diameter='pore.diameter',
    throat_cross_sectional_area='throat.cross_sectional_area'
):
    r"""
    Calculates internal surface area of pore bodies assuming they are cubes
    then subtracts the area of the neighboring throats.

    Parameters
    ----------
    target : GenericGeometry
        The Geometry object for which these values are being calculated.
        This controls the length of the calculated array, and also
        provides access to other necessary thermofluid properties.
    pore_diameter : string
        The dictionary key to the pore diameter array.
    throat_cross_sectional_area : str
        The dictionary key to the throat cross sectional area array.
        Throat areas are needed since their insection with the pore are
        removed from the computation.

    Returns
    -------
    value : NumPy ndarray
        Array containing pore surface area values.

    """
    network = target.project.network
    D = target[pore_diameter]
    Tn = network.find_neighbor_throats(pores=target.Ps, flatten=False)
    Tsurf = _np.array([network[throat_cross_sectional_area][Ts].sum() for Ts in Tn])
    value = 6 * D**2 - Tsurf
    return value


def square(
    target,
    pore_diameter='pore.diameter',
    throat_cross_sectional_area='throat.cross_sectional_area'
):
    r"""
    Calculates internal surface area of pore bodies assuming they are
    squares then subtracts the area of the neighboring throats.

    Parameters
    ----------
    target : GenericGeometry
        The Geometry object for which these values are being calculated.
        This controls the length of the calculated array, and also
        provides access to other necessary thermofluid properties.
    pore_diameter : str
        The dictionary key to the pore diameter array.
    throat_cross_sectional_area : str
        The dictionary key to the throat cross sectional area array.
        Throat areas are needed since their insection with the pore are
        removed from the computation.

    Returns
    -------
    value : NumPy ndarray
        Array containing pore surface area values.

    """
    network = target.project.network
    D = target[pore_diameter]
    Tn = network.find_neighbor_throats(pores=target.Ps, flatten=False)
    Tsurf = _np.array([network[throat_cross_sectional_area][Ts].sum() for Ts in Tn])
    value = 4 * D - Tsurf
    return value



import terminaltables as tt


class Tableist():

    def __init__(self, rows=1, cols=1, blank='---', style='ascii'):
        super().__init__()
        self.blank = blank
        header = [blank for i in range(cols)]
        self._grid = tt.AsciiTable([])
        _ = [self._grid.table_data.append(header.copy()) for _ in range(rows)]
        self._style = style

    def __getitem__(self, row):
        if isinstance(row, slice):  # If slice, convert to list
            start = row.start or 0
            stop = row.stop or self.nrows
            step = row.step or 1
            row = [r for r in range(start, stop, step)]
        if isinstance(row, list):  # If list, process each location
            temp = [self._grid.table_data[r] for r in row]
            temp2 = Tableist(len(temp), self.ncols, style=self.style)
            temp2._grid.table_data = temp
            temp2._grid.inner_heading_row_border = False
            return temp
        return self._grid.table_data[row]

    # def __setitem__(self, row, cols):
    #     if isinstance(row, int):
    #         self._grid.table_data[row] = cols
    #     if isinstance(row, slice):  # If slice, convert to list
    #         start = row.start or 0
    #         stop = row.stop or self.nrows
    #         step = row.step or 1
    #         row = [r for r in range(start, stop, step)]
    #     if isinstance(row, list):

    def set_col(self, col, vals):
        r"""
        Assign values to a column

        Parameters
        ----------
        col : int or string
            Either the column index as an integer or the column name as a
            string.
        vals
            The values to put into each row.  If a single value is given
            it is placed into each row, otherwise a list of values should be
            given containing the value for each row.

        """
        try:
            len(vals)
        except TypeError:
            vals = [vals for r in range(self.nrows)]
        for r in range(self.nrows):
            self._grid.table_data[r][col] = vals[r]

    def set_row(self, row, vals):
        r"""
        Assign values to a row

        Parameters
        ----------
        row : int or string
            Either the row index as an integer or the row name as a
            string.
        vals
            The values to put into each column.  If a single value is given
            it is placed into each column, otherwise a list of values should
            be given containing the value for each column.

        """
        try:
            len(vals)
        except TypeError:
            vals = [vals for c in range(self.ncols)]
        for c in range(self.ncols):
            self._grid.table_data[row][c] = vals[c]

    def set_row_and_col(self, row, col, val):
        r"""

        """
        header = self.get_row(0)
        index = self.get_col(0)
        try:
            col_num = header._grid.table_data[0].index(col)
        except:
            self.add_col()
            self[0][-1] = col
            col_num = -1
        try:
            row_num = index._grid.table_data.index([row])
        except:
            self.add_row()
            self[-1][0] = row
            row_num = -1
        self[row_num][col_num] = val

    def _set_style(self, style):
        if hasattr(self, '_grid'):
            data = self._grid.table_data
        else:
            data = []
        if style.lower().startswith('a'):
            self._style = 'ascii'
            self._grid = tt.AsciiTable(data)
        elif style.lower().startswith('d'):
            self._style = 'double'
            self._grid = tt.DoubleTable(data)
        elif style.lower().startswith('s'):
            self._style = 'single'
            self._grid = tt.SingleTable(data)
        elif style.lower()[0] in ['g', 'm']:
            self._style = 'markdown'
            self._grid = tt.GithubFlavoredMarkdownTable(data)
        self._grid.inner_heading_row_border = True
        self._grid.padding_left = 3
        self._grid.padding_right = 3
        self._grid.justify_columns = {col: 'center' for col in range(self.ncols)}

    def _get_style(self):
        return self._style

    style = property(fset=_set_style, fget=_get_style)

    def _set_blank(self, blank):
        if hasattr(self, '_blank'):
            for i, row in enumerate(self._grid.table_data):
                for j, col in enumerate(self._grid.table_data[i]):
                    if self._grid.table_data[i][j] == self._blank:
                        self._grid.table_data[i][j] = blank
        self._blank = blank

    def _get_blank(self):
        if hasattr(self, '_blank'):
            return self._blank

    blank = property(fset=_set_blank, fget=_get_blank)

    @property
    def ncols(self):
        header = [self.blank for i in self._grid.table_data[0]]
        return len(header)

    @property
    def nrows(self):
        return len(self._grid.table_data)

    @property
    def shape(self):
        return (self.nrows, self.ncols)

    @property
    def size(self):
        return self.nrows * self.ncols

    @property
    def nnz(self):
        nnz = 0
        for i in range(self.nrows):
            nnz += len([j for j in self.row(i) if j != self.blank])
        return nnz

    def get_row(self, row):
        r"""
        Fetch all the entries in a single row

        Parameters
        ----------
        row : int or string
            Either the row index as an integer or the row name as a
            string.

        Returns
        -------
        grid
            A grid containing only the requested row

        """
        if not isinstance(row, int):
            row = self.get_col(0)._grid.table_data.index([row])
        temp = self._grid.table_data[row]
        temp2 = Tableist(1, self.ncols)
        for col, item in enumerate(temp):
            temp2[0][col] = item
        return temp2

    def get_col(self, col):
        r"""
        Fetch all the entries in a single column

        Parameters
        ----------
        col : int or string
            Either the column index as an integer or the column name as a
            string.

        Returns
        -------
        grid
            A grid containing only the requested column

        """

        if not isinstance(col, int):
            col = self._grid.table_data[0].index(col)
        temp = [self._grid.table_data[row][col] for row in range(self.nrows)]
        temp2 = Tableist(self.nrows, 1)
        for row, item in enumerate(temp):
            temp2[row][0] = item
        return temp2

    def add_row(self, row=None, num=1):
        r"""
        Inserts blank rows below the given location

        Parameters
        ----------
        row : int
            The row below which to insert new row(s).  If not given, then
            rows are inserted at the bottom
        num : int
            The number of rows to insert.  The default is 1.

        """
        if num == 0:
            return
        header = [self.blank for i in self._grid.table_data[0]]
        if row:
            self._grid.table_data.insert(row, header)
        else:
            self._grid.table_data.append(header)
        if num > 1:
            self.add_row(row, num=num-1)

    def add_col(self, col=None, num=1):
        r"""
        Inserts blank columns beside the given location

        Parameters
        ----------
        col : int
            The column to the left of which to insert new col(s).  If not
            given, then columns are inserted on the right
        num : int
            The number of columns to insert.  The default is 1.

        """
        if num == 0:
            return
        for row in self._grid.table_data:
            if col:
                row.insert(col, self.blank)
            else:
                row.append(self.blank)
        if num > 1:
            self.add_col(col=col, num=num-1)

    def drop_col(self, col):
        r"""
        Deletes the specified column

        Parameters
        ----------
        col : int
            The column to delete

        """
        for row in self._grid.table_data:
            row.pop(col)

    def drop_row(self, row):
        r"""
        Deletes the specified row

        Parameters
        ----------
        row : int
            The row to delete

        """
        self._grid.table_data.pop(row)

    @property
    def index(self):
        return self.get_col(0)

    @property
    def header(self):
        return self.get_row(0)

    def __repr__(self):
        return self._grid.table.__str__()



import warnings
import numpy as np
import scipy as sp
import scipy.sparse as sprs
from openpnm.utils import logging, Workspace
logger = logging.getLogger(__name__)
ws = Workspace()


def find_neighbor_sites(sites, am, flatten=True, include_input=False,
                        logic='or'):
    r"""
    Given a symmetric adjacency matrix, finds all sites that are connected
    to the input sites.

    Parameters
    ----------
    am : scipy.sparse matrix
        The adjacency matrix of the network.  Must be symmetrical such that if
        sites *i* and *j* are connected, the matrix contains non-zero values
        at locations (i, j) and (j, i).

    flatten : boolean
        If ``True`` (default) the returned result is a compressed array of all
        neighbors, or a list of lists with each sub-list containing the
        neighbors for each input site.  Note that an *unflattened* list might
        be slow to generate since it is a Python ``list`` rather than a Numpy
        array.

    include_input : boolean
        If ``False`` (default) the input sites will be removed from the result.

    logic : string
        Specifies logic to filter the resulting list.  Options are:

        **'or'** : (default) All neighbors of the input sites.  This is also
        known as the 'union' in set theory or 'any' in boolean logic.  Both
        keywords are accepted and treated as 'or'.

        **'xor'** : Only neighbors of one and only one input site.  This is
        useful for finding the sites that are not shared by any of the input
        sites.  'exclusive_or' is also accepted.

        **'xnor'** : Neighbors that are shared by two or more input sites. This
        is equivalent to finding all neighbors with 'or', minus those found
        with 'xor', and is useful for finding neighbors that the inputs have
        in common.  'nxor' is also accepted.

        **'and'** : Only neighbors shared by all input sites.  This is also
        known as 'intersection' in set theory and (somtimes) as 'all' in
        boolean logic.  Both keywords are accepted and treated as 'and'.

    Returns
    -------
    An array containing the neighboring sites filtered by the given logic.  If
    ``flatten`` is ``False`` then the result is a list of lists containing the
    neighbors of each input site.

    See Also
    --------
    find_complement

    Notes
    -----
    The ``logic`` options are applied to neighboring sites only, thus it is not
    possible to obtain sites that are part of the global set but not neighbors.
    This is because (a) the list global sites might be very large, and (b) it
    is not possible to return a list of neighbors for each input site if global
    sites are considered.

    """
    if am.format != 'lil':
        am = am.tolil(copy=False)
    sites = np.array(sites, ndmin=1)
    if len(sites) == 0:
        return []
    am_coo = am.tocoo()
    n_sites = am.shape[0]
    rows = am.rows[sites].tolist()
    if len(rows) == 0:
        return []
    neighbors = am_coo.col[np.in1d(am_coo.row, sites)]
    if logic in ['or', 'union', 'any']:
        neighbors = np.unique(neighbors)
    elif logic in ['xor', 'exclusive_or']:
        neighbors = np.unique(np.where(np.bincount(neighbors) == 1)[0])
    elif logic in ['xnor', 'nxor']:
        neighbors = np.unique(np.where(np.bincount(neighbors) > 1)[0])
    elif logic in ['and', 'all', 'intersection']:
        neighbors = set(neighbors)
        [neighbors.intersection_update(i) for i in rows]
        neighbors = np.array(list(neighbors), dtype=np.int64, ndmin=1)
    else:
        raise Exception('Specified logic is not implemented')
    # Deal with removing inputs or not
    mask = np.zeros(shape=n_sites, dtype=bool)
    mask[neighbors] = True
    if not include_input:
        mask[sites] = False
    # Finally flatten or not
    if flatten:
        neighbors = np.where(mask)[0]
    else:
        if neighbors.size > 0:
            for i in range(len(rows)):
                vals = np.array(rows[i], dtype=np.int64)
                rows[i] = vals[mask[vals]]
            neighbors = rows
        else:
            neighbors = [np.array([], dtype=int) for i in range(len(sites))]
    return neighbors


def find_neighbor_bonds(sites, im=None, am=None, flatten=True, logic='or'):
    r"""
    Given an incidence matrix, finds all sites that are connected to the
    input sites.

    This function accepts either an incidence or adjacency matrix.

    Parameters
    ----------
    im : scipy.sparse matrix
        The incidence matrix of the network.  Must be shaped as (N-sites,
        N-bonds), with non-zeros indicating which sites are connected. Either
        ``am`` or ``im`` must be given.  Passing in ``im`` is slower, but more
        powerful as it allows for an unflattened list of neighbors.

    am : scipy.sparse matrix (optional)
        The adjacency matrix of the network. Either ``am`` or ``im`` must be
        given.  Passing in ``am`` is faster, but does not allow for an
        unflattened list.

    flatten : boolean (default is ``True``)
        Indicates whether the returned result is a compressed array of all
        neighbors, or a list of lists with each sub-list containing the
        neighbors for each input site.  Note that an *unflattened* list might
        be slow to generate since it is a Python ``list`` rather than a Numpy
        array.

    logic : string
        Specifies logic to filter the resulting list.  Options are:

        **'or'** : (default) All neighbors of the input sites.  This is also
        known as the 'union' in set theory or 'any' in boolean logic.  Both
        keywords are accepted and treated as 'or'.

        **'xor'** : Only neighbors of one and only one input site.  This is
        useful for finding the sites that are not shared by any of the input
        sites.  'exclusive_or' is also accepted'.

        **'xnor'** : Neighbors that are shared by two or more input sites. This
        is equivalent to finding all neighbors with 'or', minus those found
        with 'xor', and is useful for finding neighbors that the inputs have
        in common.  'nxor' is also accepted.

        **'and'** : Only neighbors shared by all input sites.  This is also
        known as 'intersection' in set theory and (somtimes) as 'all' in
        boolean logic.  Both keywords are accepted and treated as 'and'.

    Returns
    -------
    An array containing the neighboring bonds filtered by the given logic.  If
    ``flatten`` is ``False`` then the result is a list of lists containing the
    neighbors of each given input site.

    See Also
    --------
    find_complement

    Notes
    -----
    The ``logic`` options are applied to neighboring bonds only, thus it is not
    possible to obtain bonds that are part of the global set but not neighbors.
    This is because (a) the list of global bonds might be very large, and
    (b) it is not possible to return a list of neighbors for each input site
    if global sites are considered.

    """
    if im is not None:
        if im.format != 'lil':
            im = im.tolil(copy=False)
        rows = [im.rows[i] for i in np.array(sites, ndmin=1, dtype=np.int64)]
        if len(rows) == 0:
            return []
        neighbors = np.hstack(rows).astype(np.int64)
        n_bonds = int(im.nnz / 2)
        if logic in ['or', 'union', 'any']:
            neighbors = np.unique(neighbors)
        elif logic in ['xor', 'exclusive_or']:
            neighbors = np.unique(np.where(np.bincount(neighbors) == 1)[0])
        elif logic in ['xnor', 'shared']:
            neighbors = np.unique(np.where(np.bincount(neighbors) > 1)[0])
        elif logic in ['and', 'all', 'intersection']:
            neighbors = set(neighbors)
            [neighbors.intersection_update(i) for i in rows]
            neighbors = np.array(list(neighbors), dtype=int, ndmin=1)
        else:
            raise Exception('Specified logic is not implemented')
        if (flatten is False):
            if (neighbors.size > 0):
                mask = np.zeros(shape=n_bonds, dtype=bool)
                mask[neighbors] = True
                for i in range(len(rows)):
                    vals = np.array(rows[i], dtype=np.int64)
                    rows[i] = vals[mask[vals]]
                neighbors = rows
            else:
                neighbors = [np.array([], dtype=np.int64) for i in range(len(sites))]
        return neighbors
    elif am is not None:
        if am.format != 'coo':
            am = am.tocoo(copy=False)
        if not istriu(am):
            am = sp.sparse.triu(am, k=1)
        if flatten is False:
            raise Exception('flatten cannot be used with an adjacency matrix')
        Ps = np.zeros(am.shape[0], dtype=bool)
        Ps[sites] = True
        conns = np.vstack((am.row, am.col)).T
        if logic in ['or', 'union', 'any']:
            neighbors = np.any(Ps[conns], axis=1)
        elif logic in ['xor', 'exclusive_or']:
            neighbors = np.sum(Ps[conns], axis=1) == 1
        elif logic in ['xnor', 'shared']:
            neighbors = np.all(Ps[conns], axis=1)
        elif logic in ['and', 'all', 'intersection']:
            raise Exception('Specified logic is not implemented')
        else:
            raise Exception('Specified logic is not implemented')
        neighbors = np.where(neighbors)[0]
        return neighbors
    else:
        raise Exception('Either the incidence or the adjacency matrix must be specified')


def find_connected_sites(bonds, am, flatten=True, logic='or'):
    r"""
    Given an adjacency matrix, finds which sites are connected to the input
    bonds.

    Parameters
    ----------
    am : scipy.sparse matrix
        The adjacency matrix of the network.  Must be symmetrical such that if
        sites *i* and *j* are connected, the matrix contains non-zero values
        at locations (i, j) and (j, i).

    flatten : boolean (default is ``True``)
        Indicates whether the returned result is a compressed array of all
        neighbors, or a list of lists with each sub-list containing the
        neighbors for each input site.  Note that an *unflattened* list might
        be slow to generate since it is a Python ``list`` rather than a Numpy
        array.

    logic : string
        Specifies logic to filter the resulting list.  Options are:

        **'or'** : (default) All neighbors of the input bonds.  This is also
        known as the 'union' in set theory or (sometimes) 'any' in boolean
        logic.  Both keywords are accepted and treated as 'or'.

        **'xor'** : Only neighbors of one and only one input bond.  This is
        useful for finding the sites that are not shared by any of the input
        bonds.  'exclusive_or' is also accepted.

        **'xnor'** : Neighbors that are shared by two or more input bonds. This
        is equivalent to finding all neighbors with 'or', minus those found
        with 'xor', and is useful for finding neighbors that the inputs have
        in common.  'nxor' is also accepted.

        **'and'** : Only neighbors shared by all input bonds.  This is also
        known as 'intersection' in set theory and (somtimes) as 'all' in
        boolean logic.  Both keywords are accepted and treated as 'and'.

    Returns
    -------
    An array containing the connected sites, filtered by the given logic.  If
    ``flatten`` is ``False`` then the result is a list of lists containing the
    neighbors of each given input bond.  In this latter case, sites that
    have been removed by the given logic are indicated by ``nans``, thus the
    array is of type ``float`` and is not suitable for indexing.

    See Also
    --------
    find_complement

    """
    if am.format != 'coo':
        raise Exception('Adjacency matrix must be in COO format')
    bonds = np.array(bonds, ndmin=1)
    if len(bonds) == 0:
        return []
    # This function only uses the upper triangular portion, so make sure it
    # is sorted properly first
    if not istriu(am):
        am = sp.sparse.triu(am, k=1)
    neighbors = np.hstack((am.row[bonds], am.col[bonds])).astype(np.int64)
    if neighbors.size:
        n_sites = np.amax(neighbors)
    if logic in ['or', 'union', 'any']:
        neighbors = np.unique(neighbors)
    elif logic in ['xor', 'exclusive_or']:
        neighbors = np.unique(np.where(np.bincount(neighbors) == 1)[0])
    elif logic in ['xnor']:
        neighbors = np.unique(np.where(np.bincount(neighbors) > 1)[0])
    elif logic in ['and', 'all', 'intersection']:
        temp = np.vstack((am.row[bonds], am.col[bonds])).T.tolist()
        temp = [set(pair) for pair in temp]
        neighbors = temp[0]
        [neighbors.intersection_update(pair) for pair in temp[1:]]
        neighbors = np.array(list(neighbors), dtype=np.int64, ndmin=1)
    else:
        raise Exception('Specified logic is not implemented')
    if flatten is False:
        if neighbors.size:
            mask = np.zeros(shape=n_sites + 1, dtype=bool)
            mask[neighbors] = True
            temp = np.hstack((am.row[bonds], am.col[bonds])).astype(np.int64)
            temp[~mask[temp]] = -1
            inds = np.where(temp == -1)[0]
            if len(inds):
                temp = temp.astype(float)
                temp[inds] = np.nan
            temp = np.reshape(a=temp, newshape=[len(bonds), 2], order='F')
            neighbors = temp
        else:
            neighbors = [np.array([], dtype=np.int64) for i in range(len(bonds))]
    return neighbors


def find_connecting_bonds(sites, am):
    r"""
    Given pairs of sites, finds the bonds which connects each pair.

    Parameters
    ----------
    sites : array_like
        A 2-column vector containing pairs of site indices on each row.

    am : scipy.sparse matrix
        The adjacency matrix of the network.  Must be symmetrical such that if
        sites *i* and *j* are connected, the matrix contains non-zero values
        at locations (i, j) and (j, i).

    Returns
    -------
    Returns a list the same length as P1 (and P2) with each element
    containing the throat number that connects the corresponding pores,
    or `None`` if pores are not connected.

    Notes
    -----
    The returned list can be converted to an ND-array, which will convert
    the ``None`` values to ``nan``.  These can then be found using
    ``numpy.isnan``.

    """
    if am.format != 'dok':
        am = am.todok(copy=False)
    sites = np.array(sites, ndmin=2)
    if sites.size == 0:
        return []
    z = tuple(zip(sites[:, 0], sites[:, 1]))
    neighbors = [am.get(z[i], None) for i in range(len(z))]
    return neighbors


def find_complement(am, sites=None, bonds=None, asmask=False):
    r"""
    Finds the complementary sites (or bonds) to a given set of inputs

    Parameters
    ----------
    am : scipy.sparse matrix
        The adjacency matrix of the network.

    sites : array_like (optional)
        The set of sites for which the complement is sought

    bonds : array_like (optional)
        The set of bonds for which the complement is sought

    asmask : boolean
        If set to ``True`` the result is returned as a boolean mask of the
        correct length with ``True`` values indicate the complements.  The
        default is ``False`` which returns a list of indices instead.

    Returns
    -------
    An array containing indices of the sites (or bonds) that are not part of
    the input list.

    Notes
    -----
    Either ``sites`` or ``bonds`` must be specified

    """
    if (sites is not None) and (bonds is None):
        inds = np.unique(sites)
        N = am.shape[0]
    elif (bonds is not None) and (sites is None):
        inds = np.unique(bonds)
        N = int(am.nnz / 2)
    elif (bonds is not None) and (sites is not None):
        raise Exception('Only one of sites or bonds can be specified')
    else:
        raise Exception('Either sites or bonds must be specified')
    mask = np.ones(shape=N, dtype=bool)
    mask[inds] = False
    if asmask:
        return mask
    else:
        return np.arange(N)[mask]


def istriu(am):
    r"""
    Returns ``True`` is the sparse adjacency matrix is upper triangular
    """
    if am.shape[0] != am.shape[1]:
        print('Matrix is not square, triangularity is irrelevant')
        return False
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    return np.all(am.row <= am.col)


def istril(am):
    r"""
    Returns ``True`` is the sparse adjacency matrix is lower triangular
    """
    if am.shape[0] != am.shape[1]:
        print('Matrix is not square, triangularity is irrelevant')
        return False
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    return np.all(am.row >= am.col)


def istriangular(am):
    r"""
    Returns ``True`` if the sparse adjacency matrix is either upper or lower
    triangular
    """
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    return istril(am) or istriu(am)


def issymmetric(am):
    r"""
    A method to check if a square matrix is symmetric
    Returns ``True`` if the sparse adjacency matrix is symmetric
    """
    if am.shape[0] != am.shape[1]:
        logger.warning('Matrix is not square, symmetrical is irrelevant')
        return False
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    if istril(am) or istriu(am):
        return False
    # Compare am with its transpose, element wise
    sym = ((am != am.T).size) == 0
    return sym


def _am_to_im(am):
    r"""
    Convert an adjacency matrix into an incidence matrix
    """
    if am.shape[0] != am.shape[1]:
        raise Exception('Adjacency matrices must be square')
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    conn = np.vstack((am.row, am.col)).T
    row = conn[:, 0]
    data = am.data
    col = np.arange(np.size(am.data))
    if istriangular(am):
        row = np.append(row, conn[:, 1])
        data = np.append(data, data)
        col = np.append(col, col)
    im = sprs.coo.coo_matrix((data, (row, col)), (row.max() + 1, col.max() + 1))
    return im


def _im_to_am(im):
    r"""
    Convert an incidence matrix into an adjacency matrix
    """
    if im.shape[0] == im.shape[1]:
        print('Warning: Received matrix is square which is unlikely')
    if im.shape[0] > im.shape[1]:
        print('Warning: Received matrix has more sites than bonds')
    if im.format != 'coo':
        im = im.tocoo(copy=False)


def tri_to_am(tri):
    r"""
    Given a Delaunay Triangulation object from Scipy's ``spatial`` module,
    converts to a sparse adjacency matrix network representation.

    Parameters
    ----------
    tri : Delaunay Triangulation Object
        This object is produced by ``scipy.spatial.Delaunay``

    Returns
    -------
    A sparse adjacency matrix in COO format.  The network is undirected
    and unweighted, so the adjacency matrix is upper-triangular and all the
    weights are set to 1.

    """
    # Create an empty list-of-list matrix
    lil = sprs.lil_matrix((tri.npoints, tri.npoints))
    # Scan through Delaunay triangulation to retrieve pairs
    indices, indptr = tri.vertex_neighbor_vertices
    for k in range(tri.npoints):
        lil.rows[k] = indptr[indices[k]:indices[k + 1]]
    # Convert to coo format
    lil.data = lil.rows  # Just a dummy array to make things work properly
    coo = lil.tocoo()
    # Set weights to 1's
    coo.data = np.ones_like(coo.data)
    # Remove diagonal, and convert to csr remove duplicates
    am = sp.sparse.triu(A=coo, k=1, format='csr')
    # The convert back to COO and return
    am = am.tocoo()
    return am


def vor_to_am(vor):
    r"""
    Given a Voronoi tessellation object from Scipy's ``spatial`` module,
    converts to a sparse adjacency matrix network representation in COO format.

    Parameters
    ----------
    vor : Voronoi Tessellation object
        This object is produced by ``scipy.spatial.Voronoi``

    Returns
    -------
    A sparse adjacency matrix in COO format.  The network is undirected
    and unweighted, so the adjacency matrix is upper-triangular and all the
    weights are set to 1.

    """
    # Create adjacency matrix in lil format for quick matrix construction
    N = vor.vertices.shape[0]
    rc = [[], []]
    for ij in vor.ridge_dict.keys():
        row = vor.ridge_dict[ij].copy()
        # Make sure voronoi cell closes upon itself
        row.append(row[0])
        # Add connections to rc list
        rc[0].extend(row[:-1])
        rc[1].extend(row[1:])
    rc = np.vstack(rc).T
    # Make adj mat upper triangular
    rc = np.sort(rc, axis=1)
    # Remove any pairs with ends at infinity (-1)
    keep = ~np.any(rc == -1, axis=1)
    rc = rc[keep]
    data = np.ones_like(rc[:, 0])
    # Build adj mat in COO format
    M = N = np.amax(rc) + 1
    am = sprs.coo_matrix((data, (rc[:, 0], rc[:, 1])), shape=(M, N))
    # Remove diagonal, and convert to csr remove duplicates
    am = sp.sparse.triu(A=am, k=1, format='csr')
    # The convert back to COO and return
    am = am.tocoo()
    return am


def conns_to_am(conns, shape=None, force_triu=True, drop_diag=True,
                drop_dupes=True, drop_negs=True):
    r"""
    Converts a list of connections into a Scipy sparse adjacency matrix

    Parameters
    ----------
    conns : array_like, N x 2
        The list of site-to-site connections

    shape : list, optional
        The shape of the array.  If none is given then it is inferred from the
        maximum value in ``conns`` array.

    force_triu : boolean
        If True (default), then all connections are assumed undirected, and
        moved to the upper triangular portion of the array

    drop_diag : boolean
        If True (default), then connections from a site and itself are removed.

    drop_dupes : boolean
        If True (default), then all pairs of sites sharing multiple connections
        are reduced to a single connection.

    drop_negs : boolean
        If True (default), then all connections with one or both ends pointing
        to a negative number are removed.

    """
    if force_triu:  # Sort connections to [low, high]
        conns = np.sort(conns, axis=1)
    if drop_negs:  # Remove connections to -1
        keep = ~np.any(conns < 0, axis=1)
        conns = conns[keep]
    if drop_diag:  # Remove connections of [self, self]
        keep = np.where(conns[:, 0] != conns[:, 1])[0]
        conns = conns[keep]
    # Now convert to actual sparse array in COO format
    data = np.ones_like(conns[:, 0], dtype=int)
    if shape is None:
        N = conns.max() + 1
        shape = (N, N)
    am = sprs.coo_matrix((data, (conns[:, 0], conns[:, 1])), shape=shape)
    if drop_dupes:  # Convert to csr and back too coo
        am = am.tocsr()
        am = am.tocoo()
    # Perform one last check on adjacency matrix
    missing = np.where(np.bincount(conns.flatten()) == 0)[0]
    if np.size(missing) or np.any(am.col.max() < (shape[0] - 1)):
        warnings.warn('Some nodes are not connected to any bonds')
    return am



import numpy as np
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import logging, Docorator, GenericSettings
docstr = Docorator()
logger = logging.getLogger(__name__)


@docstr.get_sections(base='AdvectionDiffusionSettings',
                     sections=['Parameters', 'Other Parameters'])
@docstr.dedent
class AdvectionDiffusionSettings(GenericSettings):
    r"""
    Parameters
    ----------
    %(ReactiveTransportSettings.parameters)s
    quantity : str
        The name of the physical quantity to be calculated. The default
        value is 'pore.concentration'.
    conductance : str
        The name of the advective-diffusive conductance model to use for
        calculating the transport conductance used by the algorithm. The
        default value is 'throat.ad_dif_conductance'.
    diffusive_conductance : str
        The name of the diffusive conductance values to be used by the
        specified ``conductance`` model to find the advective-diffusive
        conductance. The default value is 'throat.diffusive_conductance'.
    hydraulic_conductance : str, optional
        The name of the hydraulic conductance values to be used by the
        specified ``conductance`` model to find the advective-diffusive
        conductance. The default value is 'throat.hydraulic_conductance'.
    pressure : str, optional
        The name of the pressure values calculated by the ``StokesFlow``
        algorithm. The default value is 'pore.pressure'.

    ----

    **The following parameters pertain to the ReactiveTransport class**

    %(ReactiveTransportSettings.other_parameters)s

    ----

    **The following parameters pertain to the GenericTransport class**

    %(GenericTransportSettings.other_parameters)s

    """

    quantity = 'pore.concentration'
    conductance = 'throat.ad_dif_conductance'
    diffusive_conductance = 'throat.diffusive_conductance'
    hydraulic_conductance = 'throat.hydraulic_conductance'
    pressure = 'pore.pressure'


class AdvectionDiffusion(ReactiveTransport):
    r"""
    A subclass of ReactiveTransport to simulate advection-diffusion
    """

    def __init__(self, settings={}, **kwargs):
        super().__init__(**kwargs)
        self.settings._update_settings_and_docs(AdvectionDiffusionSettings())
        self.settings.update(settings)

    def setup(
            self,
            phase=None,
            quantity='',
            conductance='',
            diffusive_conductance='',
            hydraulic_conductance='',
            pressure='',
            **kwargs
    ):
        r"""
        Setup method for setting/modifying algorithm settings
        """
        if phase:
            self.settings['phase'] = phase.name
        if quantity:
            self.settings['quantity'] = quantity
        if conductance:
            self.settings['conductance'] = conductance
        if diffusive_conductance:
            self.settings['diffusive_conductance'] = diffusive_conductance
        if hydraulic_conductance:
            self.settings['hydraulic_conductance'] = hydraulic_conductance
        if pressure:
            self.settings['pressure'] = pressure
        super().setup(**kwargs)

    def set_outflow_BC(self, pores, mode='merge'):
        r"""
        Adds outflow boundary condition to the selected pores

        Parameters
        ----------
        pores : array_like
            The pore indices where the condition should be applied
        mode : string, optional
            Controls how the boundary conditions are applied.  Options are:

            'merge' - (Default) Adds supplied boundary conditions to already
            existing conditions, and also overwrites any existing values.
            If at rate or value BC exists at the given locations, these
            are deleted, and outflow conditions are given priority.
            'overwrite' - Deletes all boundary conditions of the given type
            then adds the specified new ones.

        Notes
        -----
        Outflow condition means that the gradient of the solved quantity
        does not change, i.e. is 0.

        """
        # Hijack the parse_mode function to verify mode/pores argument
        mode = self._parse_mode(mode, allowed=['merge', 'overwrite'],
                                single=True)
        pores = self._parse_indices(pores)

        # Calculating A[i,i] values to ensure the outflow condition
        network = self.project.network
        phase = self.project.phases()[self.settings['phase']]
        throats = network.find_neighbor_throats(pores=pores)
        C12 = network.conns[throats]
        P12 = phase[self.settings['pressure']][C12]
        gh = phase[self.settings['hydraulic_conductance']][throats]
        Q12 = -gh * np.diff(P12, axis=1).squeeze()
        Qp = np.zeros(self.Np)
        np.add.at(Qp, C12[:, 0], -Q12)
        np.add.at(Qp, C12[:, 1], Q12)

        # Ensure other BCs are not already applied at given pores
        hits = ~np.isnan(self['pore.bc_rate'][pores])
        if np.any(hits):
            self['pore.bc_rate'][pores] = np.nan
            logger.warning('Rate boundary conditions found in some of the '
                           + 'specified pores will be overwritten')
        hits = ~np.isnan(self['pore.bc_value'][pores])
        if np.any(hits):
            self['pore.bc_value'][pores] = np.nan
            logger.warning('Value boundary conditions found in some of the '
                           + 'specified pores will be overwritten')
        # Store boundary values
        if ('pore.bc_outflow' not in self.keys()) or (mode == 'overwrite'):
            self['pore.bc_outflow'] = np.nan
        self['pore.bc_outflow'][pores] = Qp[pores]

    def remove_BC(self, pores=None, bctype='all'):
        # parse bctype argument
        if isinstance(bctype, str):
            bctype = [bctype]
        if 'all' in bctype:
            bctype = ['value', 'rate', 'outflow']
        if ('pore.bc_outflow' in self.keys()) and ('outflow' in bctype):
            self['pore.bc_outflow'][pores] = np.nan
        super().remove_BC(pores=pores, bctype=bctype)

    def _apply_BCs(self):
        r"""
        Applies Dirichlet, Neumann, and outflow BCs in order
        """
        # Apply Dirichlet and rate BCs
        ReactiveTransport._apply_BCs(self)
        if 'pore.bc_outflow' not in self.keys():
            return
        # Apply outflow BC
        diag = self.A.diagonal()
        ind = np.isfinite(self['pore.bc_outflow'])
        diag[ind] += self['pore.bc_outflow'][ind]
        self.A.setdiag(diag)

    def _set_BC(self, pores, bctype, bcvalues=None, mode='merge'):
        pores = self._parse_indices(pores)
        # First check that given pores outflow BCs already applied
        if 'pore.bc_outflow' in self.keys():
            hits = ~np.isnan(self['pore.bc_outflow'][pores])
            if np.any(hits):
                raise Exception('Cannot apply BCs to the following pores '
                                + 'which already have an outflow BC '
                                + 'specified', pores[np.where(hits)])
        # Then call parent class function if above check passes
        super()._set_BC(pores=pores, bctype=bctype, bcvalues=bcvalues, mode=mode)



import json
import json_tricks as jsont
import numpy as np
import importlib
from datetime import datetime
from openpnm.utils import Workspace, Project
from openpnm.utils import logging
from openpnm.io import GenericIO
from h5py import File as hdfFile
logger = logging.getLogger(__name__)
ws = Workspace()


class PNM(GenericIO):
    r"""
    This is the official way to save and load OpenPNM projects
    """
    @classmethod
    def save_project(cls, project, filename=None):
        if filename is None:
            filename = project.name + '.pnm'

        # Make a directory using the given file name
        f = cls._parse_filename(filename, 'pnm')
        with hdfFile(f, mode='w') as root:
            # root = hdfFile(f, mode='w')
            root.attrs['version'] = ws.version
            date = datetime.today().strftime("%Y %h %d %H:%M:%S")
            root.attrs['date saved'] = date
            root.attrs['name'] = project.name
            # root.attrs['comments'] = project.comments
            for obj in project:
                found_attrs = set(obj.__dict__.keys())
                known_attrs = set(['settings', '_models_dict',
                                   '_am', '_im',
                                   '_spacing', '_shape'])
                foreign_attrs = found_attrs.difference(known_attrs)
                if len(foreign_attrs) > 0:
                    line_break = f"\n{'':13}"
                    logger.critical(f"{obj.name} has the following attributes that will"
                                    + f" not be saved: {[i for i in foreign_attrs]}"
                                    + f"{line_break}Consider using Pickle instead")
                item = root.create_group(obj.name)
                for arr in obj.keys():  # Store data
                    try:
                        item.create_dataset(name=arr, data=obj[arr],
                                            shape=obj[arr].shape,
                                            compression="gzip")
                    except TypeError:  # Deal with 'object' arrays
                        logger.warning(arr + ' is being converted to a string')
                        b = jsont.dumps(obj[arr])
                        c = b.encode()
                        d = np.void(c)
                        item.create_dataset(name=arr, data=d)
                # Store settings dict as metadata
                item.attrs['settings'] = json.dumps(obj.settings)
                # Store models dict as metadata
                if hasattr(obj, 'models'):
                    obj_models = {}
                    for model in obj.models.keys():
                        temp = {k: v for k, v in obj.models[model].items()
                                if k != 'model'}
                        if 'model' in obj.models[model].keys():
                            a = obj.models[model]['model']
                            temp['model'] = a.__module__ + '|' + \
                                a.__code__.co_name
                        obj_models[model] = temp
                    try:
                        item.attrs['models'] = json.dumps(obj_models)
                    except TypeError:
                        logger.critical('The model ' + model + ' and it\'s '
                                        + 'parameters could not be written '
                                        + 'to file')
                item.attrs['class'] = str(obj.__class__)

    @classmethod
    def load_project(cls, filename):
        f = cls._parse_filename(filename, 'pnm')
        with hdfFile(f, mode='r') as root:
            logger.info('Loading project from file ' + f.name)
            try:  # Create an empty project with old name
                proj = Project(name=root.attrs['name'])
                logger.info('Loading ' + proj.name)
            except Exception:  # Generate a new name if collision occurs
                proj = Project()
                logger.warning('A project named ' + root.attrs['name']
                               + ' already exists, renaming to ' + proj.name)
            logger.info('Created using OpenPNM version ' + root.attrs['version'])
            logger.info('Saved on ' + root.attrs['date saved'])
            loglevel = ws.settings['loglevel']
            ws.settings['loglevel'] = 50
            for name in root.keys():
                if 'network' in root[name].attrs['class']:
                    proj, obj = create_obj(root, name, proj)
            for name in root.keys():
                if 'network' not in root[name].attrs['class']:
                    proj, obj = create_obj(root, name, proj)
            ws.settings['loglevel'] = loglevel
        return proj


def create_obj(root, name, proj):
    r"""
    Reproduces an OpenPNM object, given the hdf5 file and name
    """
    import openpnm as op
    # regenerate object as same class
    mro = root[name].attrs['class']
    mro = mro.split("'")[1]
    mro = mro.split('.')
    mod = importlib.import_module('.'.join(mro[:-1]))
    clss = getattr(mod, mro[-1])
    obj = clss.__new__(cls=clss)
    obj._name = name
    # Add data to obj
    for arr in root[name].keys():
        a = np.array(root[name][arr])
        if str(a.dtype).startswith("|V"):
            logger.warning(arr + ' is being converted from string')
            b = np.string_(a)
            c = b.astype(str)
            a = jsont.loads(c)
        obj.update({arr: a})
    # Add settings to obj
    obj.settings.update(json.loads(root[name].attrs['settings']))
    # Add models to obj
    if hasattr(obj, 'models'):
        obj.models.update(json.loads(root[name].attrs['models']))
        for m in obj.models.keys():
            md, fn = obj.models[m]['model'].split('|')
            try:
                md = importlib.import_module(md)
                try:
                    obj.models[m]['model'] = getattr(md, fn)
                except AttributeError:
                    logger.warning(f"The function {fn} could not be loaded, adding"
                                   + " 'blank' instead")
                    obj.models[m]['model'] = op.models.misc.basic_math.blank
            except ModuleNotFoundError:
                logger.warning(f"The module {md} could not be loaded, adding"
                               + " 'blank' instead")
                obj.models[m]['model'] = op.models.misc.basic_math.blank
    proj.append(obj)
    return proj, obj



r"""
This submodule contains pore-scale models that calculate geometrical
properties. These models are to be added to a Geometry object.
"""

from . import pore_size
from . import pore_seed
from . import pore_volume
from . import pore_surface_area
from . import pore_cross_sectional_area
from . import throat_endpoints
from . import throat_cross_sectional_area
from . import throat_seed
from . import throat_size
from . import throat_length
from . import throat_perimeter
from . import throat_surface_area
from . import throat_volume
from . import throat_capillary_shape_factor
from . import throat_centroid
from . import throat_vector

# Up for deprecation
pore_area = pore_cross_sectional_area
throat_area = throat_cross_sectional_area



import numpy as np
import openpnm.models.misc as misc
from openpnm.phases import GenericPhase as GenericPhase
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class MultiPhase(GenericPhase):
    r"""
    Creates Phase object that represents a multiphase system consisting of
    a given list of OpenPNM Phase objects.

    Parameters
    ----------
    network : GenericNetwork
        The network to which this phase object will be attached.
    project : Project, optional
        The Project with which this phase should be associted. If a
        ``network`` is given then this is ignored and the Network's
        project is used. If a ``network`` is not given then this is
        mandatory.
    name : str, optional
        The name of the phase. This is useful to keep track of the objects
        throughout the simulation. The name must be unique to the project.
        If no name is given, one is generated.

    Examples
    --------
    >>> import scipy as sp
    >>> import openpnm as op
    >>> from openpnm.phases import Air, Water, MultiPhase

    >>> net = op.network.Cubic(shape=[5, 5, 5])
    >>> air = Air(network=net, name='air')  # Create two pure phases
    >>> water = Water(network=net, name='water')
    >>> mphase = MultiPhase(network=net, phases=[air, water], name='multi')
    >>> Ps = net['pore.coords'][:, 0] < 3  # Pick some pores to be air filled
    >>> Ts = net.find_neighbor_throats(pores=Ps)  # Find neighboring throats
    >>> Ts = net.tomask(throats=Ts)  # Convert throat indices to mask
    >>> mphase.set_occupancy(phase=air, Pvals=Ps, Tvals=Ts)  # Assign occupancies
    >>> mphase.set_occupancy(phase=water, Pvals=~Ps, Tvals=~Ts)

    Air and water have uniform viscosity values throughout, so the
    peak-to-peak distance is 0, while the mixture phase has the viscosity
    of air in some locations and water in others, hence a heterogenous
    viscosity array:

    >>> np.ptp(water['pore.viscosity']) == 0.
    True
    >>> np.ptp(air['pore.viscosity']) == 0.
    True
    >>> np.ptp(mphase['pore.viscosity']) == 0.
    False

    """

    def __init__(self, phases=[], settings={}, **kwargs):
        super().__init__(**kwargs)
        self.settings.update(
            {
                'phases': [],
                'throat_occupancy': 'manual',
                'partition_coef_prefix': 'throat.partition_coef'
            }
        )
        self.settings.update(settings)

        self['pore.occupancy.all'] = np.zeros(self.Np, dtype=float)
        self['throat.occupancy.all'] = np.zeros(self.Nt, dtype=float)

        # Pressure/temperature must come from constituent phases
        self.pop('pore.temperature', None)
        self.pop('pore.pressure', None)

        # Add a dummy model for assembling the global parition
        # coefficients array. The dummy model only accepts 'target' and
        # kwargs. When a binary partition coefficient is added to the
        # MultiPhase, we manually append the binary partition coefficient
        # propname as a keyword argument to this dummy model, so that our
        # dependency handler is notified of this dependency!
        partition_coef_global = self.settings['partition_coef_prefix'] + ".all"
        self.add_model(propname=partition_coef_global, model=_dummy)

        # Add supplied phases to phases dict and initialize occupancy to 0
        self.add_phases(phases)

        logger.warning('MultiPhase is a beta feature. Functionality may change!')

    def __getitem__(self, key):
        try:
            vals = super().__getitem__(key)
        except KeyError:
            vals = self.interleave_data(key)
        return vals

    def _get_phases(self):
        phases = {self.project[item].name: self.project[item]
                  for item in self.settings['phases']}
        return phases

    phases = property(fget=_get_phases)

    def _update_occupancy(self):
        r"""Updates 'occupancy.all' by summing up all occupancies"""
        for elem in ["pore", "throat"]:
            dict_ = self[f"{elem}.occupancy"]
            dict_.pop(f"{elem}.occupancy.all")
            self[f"{elem}.occupancy.all"] = np.sum(list(dict_.values()), axis=0)

    def add_phases(self, phases):
        r"""
        Adds supplied phases to the MultiPhase object and initializes
        occupancy to 0. It does not return a value.

        Parameters
        ----------
        phases : List[GenericPhase]

        """
        phases = np.array(phases, ndmin=1)
        for phase in phases:
            if phase.name not in self.settings['phases']:
                self.settings['phases'].append(phase.name)
                self[f'pore.occupancy.{phase.name}'] = 0.
                self[f'throat.occupancy.{phase.name}'] = 0.

    def set_binary_partition_coef(self, phases, model, **kwargs):
        r"""
        Sets binary partition coefficient as defined by the interface
        concentration ratio of phase 1 to phase 2.

        Parameters
        ----------
        phases : List[GenericPhase]
            List of the two phases for which the binary partition
            coefficient model is being added.
        model : OpenPNM model
            Model for calculating the binary partition coefficient.
        kwargs : dict
            Keyword arguments to be passed to the ``model``.

        Example
        -------
        >>> import openpnm as op
        >>> from openpnm.phases import Air, Water, MultiPhase
        >>> from openpnm.models.misc import constant

        >>> net = op.network.Cubic(shape=[5, 5, 5])
        >>> air = Air(network=net, name='air')  # Create two pure phases
        >>> water = Water(network=net, name='water')
        >>> mphase = MultiPhase(network=net, phases=[air, water], name='multi')

        Now, assign some pores to air and the rest to water

        >>> Ps = net['pore.coords'][:, 0] < 3  # Pick some pores to be air filled
        >>> Ts = net.find_neighbor_throats(pores=Ps)  # Find neighboring throats
        >>> Ts = net.tomask(throats=Ts)  # Convert throat indices to mask
        >>> mphase.set_occupancy(phase=air, Pvals=Ps, Tvals=Ts)  # Assign occupancies
        >>> mphase.set_occupancy(phase=water, Pvals=~Ps, Tvals=~Ts)

        Now, add an interface model for binary partition coefficient

        >>> mphase.set_binary_partition_coef(phases=[air, water],
        ...                                  model=constant,
        ...                                  value=0.5)

        Now, verify that K12 for interface throats is 0.5

        >>> Ts_interface = net.find_neighbor_throats(Ps, mode="xor")
        >>> K12_interface = mphase["throat.partition_coef.all"][Ts_interface]
        >>> assert K12_interface.mean() == 0.5

        Finally, verify that K12 for non-interface throats is 1.

        >>> Ts_rest = ~ np.isin(net.Ts, Ts_interface)
        >>> K12_rest = mphase["throat.partition_coef.all"][Ts_rest]
        >>> assert K12_rest.mean() == 1.

        """
        if np.size(phases) != 2:
            raise Exception("'phases' must contain exactly two elements!")

        # Add partition coefficient interface model to the MultiPhase
        propname_prefix = self.settings["partition_coef_prefix"]
        self._add_interface_prop(propname_prefix, phases, model, **kwargs)

        # Update global partition coef. model's args (for dependency handling)
        partition_coef_global = self.settings['partition_coef_prefix'] + ".all"
        K_global_model = self.models[partition_coef_global]
        # Append binary partition coefficient propname to model args
        propname = self._format_interface_prop(propname_prefix, phases)
        key = f"K_{phases[0].name}_{phases[1].name}"
        K_global_model[key] = propname

        # Regenerate 'throat.parition_coef.all'
        self.regenerate_models(partition_coef_global)

    def _add_interface_prop(self, propname, phases, model, **kwargs):
        r"""
        Helper method used to add interface models to the ``MultiPhase``
        object by augmenting.

        Notes
        -----
        For convenience and as an OpenPNM convention, the ``propname`` is
        augmented as outlined by the following example.

        Example
        -------
        Let's say the two phases corresponding to the interface model are
        named: 'air' and 'water', and the interface propname to be added
        is 'throat.foo'. After augmentation, 'throat.foo.air:water' will
        be the propname that's stored on the ``MultiPhase`` object.

        Note that because of this convention, the order of the phases that
        are passed to this method is important.

        """
        # Add "throat" keyword to the begining of propname if no identifier is found
        if propname.split(".")[0] not in ["pore", "throat"]:
            propname = f"throat.{propname}"
        # Check propname is throat property
        if propname.startswith("pore"):
            raise Exception("'propname' must be a throat property")
        # Add model to Multiphase
        propname = self._format_interface_prop(propname, phases)
        self.add_model(propname, model, **kwargs)

    def _format_interface_prop(self, propname, phases):
        r"""
        Returns formatted interface propname.

        Parameters
        ----------
        propname : str
            Dictionary key of the interface property.
        phases : List[GenericPhase]
            List of two phases that together form the interface property.

        Example
        -------
        >>> import openpnm as op
        >>> net = op.network.Cubic(shape=[5, 5, 5])
        >>> air = op.phases.Air(network=net, name='air')
        >>> water = op.phases.Water(network=net, name='water')
        >>> mphase = op.phases.MultiPhase(network=net, phases=[air, water])
        >>> mphase._format_interface_prop(propname="throat.foo", phases=[air, water])
        'throat.foo.air:water'

        """
        prefix = propname
        suffix = ":".join(phase.name for phase in phases)
        return f"{prefix}.{suffix}"

    def _get_phases_names(self, formatted_propname):
        r"""Retrieves phases' names from a formatted propname"""
        return formatted_propname.split(".")[2].split(":")

    def _assemble_partition_coef_global(self):
        r"""Updates the global partition coefficient array"""
        conns = self.network.conns
        partition_coef_models = []
        prefix = self.settings["partition_coef_prefix"]
        K_global = np.ones(self.Nt)
        partition_coef_global = prefix + ".all"

        # Find all binary partition coefficient models
        for model_key in self.models.keys():
            if model_key.startswith(prefix):
                if model_key != partition_coef_global:
                    partition_coef_models.append(model_key)

        # Modify the global partition coefficient for each phase pair
        for propname in partition_coef_models:
            K12 = self[propname]
            phases_names = self._get_phases_names(propname)
            S1 = self[f"pore.occupancy.{phases_names[0]}"][conns]
            S2 = self[f"pore.occupancy.{phases_names[1]}"][conns]
            mask = (S1[:, 0] + S2[:, 1]) == 2.
            K_global[mask] = K12[mask]
            mask = (S2[:, 0] + S1[:, 1]) == 2.
            K_global[mask] = 1 / K12[mask]

        return K_global

    def interleave_data(self, prop):
        r"""
        Gathers property values from component phases to build a single
        array.

        If the requested ``prop`` is not on this MultiPhase, then a search
        is conducted on all associated phase objects, and values from each
        are assembled into a single array.

        Parameters
        ----------
        prop : str
            The property to be retrieved.

        Returns
        -------
        ndarray
            An array containing the specified property retrieved from each
            component phase and assembled based on the specified mixing
            rule.

        """
        element = self._parse_element(prop)[0]
        vals = np.zeros([self._count(element=element)], dtype=float)
        # Retrieve property from constituent phases (weight = occupancy)
        try:
            for phase in self.phases.values():
                vals += phase[prop] * self[f"{element}.occupancy.{phase.name}"]
        # Otherwise - if not found - retrieve from super class
        except KeyError:
            vals = super().interleave_data(prop)

        # Check for consistency of occupancy values (i.e. add up to 1)
        if np.any(self[f"{element}.occupancy.all"] != 1.0):
            self._update_occupancy()
            if np.any(self[f"{element}.occupancy.all"] != 1.0):
                raise Exception(f"Occupancy doesn't add to unity in all {element}s")

        return vals

    def regenerate_models(self, propnames=None, exclude=[], deep=False):
        r"""
        Regenerate models associated with the Multiphase object

        This method works by first regenerating the models associated with
        the constituent phases, and then regenerating Multiphase models.

        Parameters
        ----------
        propnames : List[str] or str
            The list of property names to be regenerated. If None are
            given then ALL models are re-run (except for those whose
            ``regen_mode`` is 'constant').
        exclude : List[str]
            Since the default behavior is to run ALL models, this can be
            used to exclude specific models. It may be more convenient to
            supply as list of 2 models to exclude than to specify 8 models
            to include.
        deep : bool
            Specifies whether or not to regenerate models on all
            associated objects. For instance, if ``True``, then all
            Physics models will be regenerated when method is called on
            the corresponding Phase. The default is ``False``. The method
            does not work in reverse, so regenerating models on a Physics
            will not update a Phase.

        """
        # Regenerate models associated with phases within MultiPhase object
        for phase in self.phases.values():
            phase.regenerate_models(propnames=propnames, exclude=exclude, deep=deep)
        # Regenerate models specific to MultiPhase object
        super().regenerate_models(propnames=propnames, exclude=exclude, deep=deep)

    def _set_automatic_throat_occupancy(self, mode="mean"):
        r"""
        Automatically interpolates throat occupancy based on that in
        adjacent pores. This method doesn't return any value.

        Parameters
        ----------
        mode : str
            Interpolation method, ex. 'mean' sets the throat occupancy as
            the average of that in adjacent pores, while 'min' sets it to
            the minimum of the two. Options are 'mean', 'min', and 'max'.

        """
        self.settings['throat_occupancy'] = 'automatic'
        for phase in self.phases.values():
            self.add_model(
                propname=f"throat.occupancy.{phase.name}",
                model=misc.from_neighbor_pores,
                prop=f"pore.occupancy.{phase.name}",
                mode=mode
            )

    def set_occupancy(self, phase, Pvals=[], Tvals=[], pores=[], throats=[]):
        r"""
        Specifies occupancy of a phase in each pore and/or throat. This
        method doesn't return any value.

        Parameters
        ----------
        phase : GenericPhase
            The phase whose occupancy is being specified.
        Pvals : array_like
            The volume fraction of ``phase`` in each pore. This array must
            be Np-long, except when ``pores`` is also specified, where in
            that case they must be of equal length. If a scalar is
            received, it is applied to all the pores. If nothing is passed,
            ``Pvals=1.0`` will be assumed.
        Tvals : array_like
            The volume fraction of ``phase`` in each throat. This array
            must be Nt-long, except when ``throats`` is also specified,
            where in that case they must be of equal length. If a scalar
            is received it is applied to all the throats. If nothing is
            passed, ``Tvals=1.0`` will be assumed.
        pores : array_like
            The location of pores whose occupancy is to be set.
        throats : array_like
            The location of throats whose occupancy is to be set.

        """
        # TODO: pores/throats could also be masks

        Pvals = np.array(Pvals, ndmin=1)
        Tvals = np.array(Tvals, ndmin=1)
        pores = np.array(pores, ndmin=1)
        throats = np.array(throats, ndmin=1)

        # Check for size consistency of the arguments
        if Pvals.size and pores.size:
            if Pvals.size != 1 and pores.size != 1:
                if Pvals.size != pores.size:
                    raise Exception("Pvals and pores must be the same size.")
        if Tvals.size and throats.size:
            if Tvals.size != 1 and throats.size != 1:
                if Tvals.size != throats.size:
                    raise Exception("Tvals and throats must be the same size.")

        # Check if the passed phase is already part of MultiPhase object
        if phase not in self.project:
            raise Exception(f"{phase.name} doesn't belong to this project")
        # Add the passed phase to MultiPhase object if not found
        else:
            if phase.name not in self.settings['phases']:
                self.add_phases(phase)

        # Check for value consistency of the arguments
        if np.any(Pvals > 1.0) or np.any(Pvals < 0.0):
            logger.critical('Received Pvals contain volume fractions outside '
                            + 'the range of 0 to 1')
        if np.any(Tvals > 1.0) or np.any(Tvals < 0.0):
            logger.critical('Received Tvals contain volume fractions outside '
                            + 'the range of 0 to 1')

        if Pvals.size and not pores.size:
            pores = self.pores()
        if Tvals.size and not throats.size:
            throats = self.throats()

        if pores.size:
            Pvals = Pvals if Pvals.size else 1.0
            self['pore.occupancy.' + phase.name][pores] = Pvals
        if throats.size:
            Tvals = Tvals if Tvals.size else 1.0
            self['throat.occupancy.' + phase.name][throats] = Tvals

        if self.settings["throat_occupancy"] == "automatic":
            self.regenerate_models(propnames=f"throat.occupancy.{phase.name}")

        self._update_occupancy()


def _dummy(target, **kwargs):
    r"""Dummy model to store propnames as kwargs for dependency handling"""
    return target._assemble_partition_coef_global()



r"""
Pore-scale models to be used as source term (e.g. reaction) in algorithms.
"""
import numpy as _np
import scipy as _sp

__all__ = [
    "charge_conservation",
    "standard_kinetics",
    "linear",
    "power_law",
    "exponential",
    "natural_exponential",
    "logarithm",
    "natural_logarithm",
    "linear_sym",
    "power_law_sym",
    "exponential_sym",
    "natural_exponential_sym",
    "logarithm_sym",
    "natural_logarithm_sym",
    "general_symbolic",
    "butler_volmer_conc",
    "butler_volmer_voltage"
]


def charge_conservation(target, phase, p_alg, e_alg, assumption):
    r"""
    Applies the source term on the charge conservation equation when solving
    for ions transport.

    Parameters
    ----------
    phase : OpenPNM Phase object
            The phase on which the charge conservation equation is applied.

    p_alg : OpenPNM Algorithm object
            The algorithm used to enforce charge conservation.

    e_alg : list of OpenPNM algorithms
            The list of algorithms used to solve for transport of different
            ionic species of the mixture phase.

    assumption : string
            A string correponding to the assumption adopted to enforce charge
            conservation.

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function for the given list
                     of algortihms under the provided assumption.

        **'S1'** - A placeholder (zero array).

        **'S2'** - The value of the source term function for the given list of
                   algortihms under the provided assumption (same as 'rate').

    Notes
    -----
    Three assumptions are supported; "poisson", "electroneutrality" and
    "laplace".

    """
    import scipy.sparse.csgraph as _spgr

    F = 96485.3321233100184
    rhs = _np.zeros(shape=(p_alg.Np, ), dtype=float)
    network = p_alg.project.network
    if assumption == 'poisson':
        v = network['pore.volume']
        for e in e_alg:
            rhs += (v * F * phase['pore.valence.'+e.settings['ion']]
                    * e[e.settings['quantity']])
    elif assumption == 'poisson_2D':
        s = network['pore.area']
        for e in e_alg:
            rhs += (s * F * phase['pore.valence.'+e.settings['ion']]
                    * e[e.settings['quantity']])
    elif assumption in ['electroneutrality', 'electroneutrality_2D']:
        for e in e_alg:
            try:
                c = e[e.settings['quantity']]
            except KeyError:
                c = _np.zeros(shape=(e.Np, ), dtype=float)
            network = e.project.network
            g = phase['throat.diffusive_conductance.'+e.settings['ion']]
            am = network.create_adjacency_matrix(weights=g, fmt='coo')
            A = _spgr.laplacian(am)
            rhs += - F * phase['pore.valence.'+e.settings['ion']] * (A * c)
    elif assumption in ['laplace', 'laplace_2D']:
        pass  # rhs should remain 0
    else:
        raise Exception('Unknown keyword for "charge_conservation", can '
                        + 'only be "poisson", "poisson_2D", "laplace", '
                        + '"laplace_2D", "electroneutrality" or '
                        + "electroneutrality_2D")
    S1 = _np.zeros(shape=(p_alg.Np, ), dtype=float)
    values = {'S1': S1, 'S2': rhs, 'rate': rhs}
    return values


def standard_kinetics(target, X, prefactor, exponent):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of ``X``:

        .. math::
            r = A X^b

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    prefactor, exponent : {string}
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    quantity : {string}
        The dictionary key on the target object containing the the quantity
        of interest

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    X = target[X]
    A = target[prefactor]
    b = target[exponent]

    r = A*(X**b)
    S1 = A*b*(X**(b - 1))
    S2 = A*(1 - b)*(X**b)
    values = {'S1': S1, 'S2': S2, 'rate': r}
    return values


def _parse_args(target, key, default):
    r"""
    Parses the passed ``key``: returns ``default`` if ``key`` is empty,
    otherwise fetches and returns the values from ``target``.
    """
    if key == '':
        val = default
    else:
        val = target[key]
    return val


def linear(target, X, A1='', A2=''):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of ``X``:

        .. math::
            r = A_{1} X + A_{2}

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    A1 -> A2 : string
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    A = _parse_args(target=target, key=A1, default=0.0)
    B = _parse_args(target=target, key=A2, default=0.0)
    X = target[X]

    r = A * X + B
    S1 = A
    S2 = B
    values = {'S1': S1, 'S2': S2, 'rate': r}
    return values


def power_law(target, X, A1='', A2='', A3=''):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of *X*:

        .. math::
            r = A_{1} x^{A_{2}} + A_{3}

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    A1 -> A3 : string
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    A = _parse_args(target=target, key=A1, default=0.0)
    B = _parse_args(target=target, key=A2, default=1.0)
    C = _parse_args(target=target, key=A3, default=0.0)
    X = target[X]

    r = A * X ** B + C
    S1 = A * B * X ** (B - 1)
    S2 = A * X ** B * (1 - B) + C
    values = {'S1': S1, 'S2': S2, 'rate': r}
    return values


def exponential(target, X, A1='', A2='', A3='', A4='', A5='', A6=''):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of `X`:

        .. math::
            r =  A_{1} A_{2}^{( A_{3} x^{ A_{4} } + A_{5})} + A_{6}

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    A1 -> A6 : string
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    A = _parse_args(target=target, key=A1, default=0.0)
    B = _parse_args(target=target, key=A2, default=1.0)
    C = _parse_args(target=target, key=A3, default=1.0)
    D = _parse_args(target=target, key=A4, default=1.0)
    E = _parse_args(target=target, key=A5, default=0.0)
    F = _parse_args(target=target, key=A6, default=0.0)
    X = target[X]

    r = A * B ** (C * X ** D + E) + F
    S1 = A * C * D * X ** (D - 1) * _np.log(B) * B ** (C * X ** D + E)
    S2 = A * B ** (C * X ** D + E) * (1 - C * D * _np.log(B) * X ** D) + F
    values = {'S1': S1, 'S2': S2, 'rate': r}
    return values


def natural_exponential(target, X, A1='', A2='', A3='', A4='', A5=''):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of `X`:

        .. math::
            r = A_{1} exp( A_{2}  x^{ A_{3} } + A_{4} )+ A_{5}

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    A1 -> A5 : string
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    A = _parse_args(target=target, key=A1, default=0.0)
    B = _parse_args(target=target, key=A2, default=0.0)
    C = _parse_args(target=target, key=A3, default=0.0)
    D = _parse_args(target=target, key=A4, default=0.0)
    E = _parse_args(target=target, key=A5, default=0.0)
    X = target[X]

    r = A * _np.exp(B * X ** C + D) + E
    S1 = A * B * C * X ** (C - 1) * _np.exp(B * X ** C + D)
    S2 = A * (1 - B * C * X ** C) * _np.exp(B * X ** C + D) + E
    values = {'pore.S1': S1, 'pore.S2': S2, 'pore.rate': r}
    return values


def logarithm(target, X, A1='', A2='', A3='', A4='', A5='', A6=''):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of `X`:

        .. math::
            r =  A_{1} Log_{ A_{2} }( A_{3} x^{ A_{4} }+ A_{5})+ A_{6}

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    A1 -> A6 : string
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    A = _parse_args(target=target, key=A1, default=0.0)
    B = _parse_args(target=target, key=A2, default=10.0)
    C = _parse_args(target=target, key=A3, default=1.0)
    D = _parse_args(target=target, key=A4, default=1.0)
    E = _parse_args(target=target, key=A5, default=0.0)
    F = _parse_args(target=target, key=A6, default=0.0)
    X = target[X]

    r = (A * _np.log(C * X ** D + E)/_np.log(B) + F)
    S1 = A * C * D * X ** (D - 1) / (_np.log(B) * (C * X ** D + E))
    S2 = A * _np.log(C * X ** D + E) / _np.log(B) + F - A * C * D * X ** D / \
        (_np.log(B) * (C * X ** D + E))
    values = {'S1': S1, 'S2': S2, 'rate': r}
    return values


def natural_logarithm(target, X, A1='', A2='', A3='', A4='', A5=''):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of `X`:

        .. math::
            r =   A_{1} Ln( A_{2} x^{ A_{3} }+ A_{4})+ A_{5}

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    A1 -> A5 : string
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    A = _parse_args(target=target, key=A1, default=0.0)
    B = _parse_args(target=target, key=A2, default=1.0)
    C = _parse_args(target=target, key=A3, default=1.0)
    D = _parse_args(target=target, key=A4, default=0.0)
    E = _parse_args(target=target, key=A5, default=0.0)
    X = target[X]

    r = A*_np.log(B*X**C + D) + E
    S1 = A*B*C*X**(C - 1) / (B * X ** C + D)
    S2 = A*_np.log(B*X**C + D) + E - A*B*C*X**C / (B*X**C + D)
    values = {'pore.S1': S1, 'pore.S2': S2, 'pore.rate': r}
    return values


def _build_func(eq, **args):
    r'''
    Take a symbolic equation and return the lambdified version plus the
    linearization of form S1 * x + S2
    '''
    from sympy import lambdify
    eq_prime = eq.diff(args['x'])
    s1 = eq_prime
    s2 = eq - eq_prime*args['x']
    EQ = lambdify(args.values(), expr=eq, modules='numpy')
    S1 = lambdify(args.values(), expr=s1, modules='numpy')
    S2 = lambdify(args.values(), expr=s2, modules='numpy')
    return EQ, S1, S2


def linear_sym(target, X, A1='', A2=''):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of *x*:

        .. math::
            r = A_{1} x + A_{2}

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    A1 -> A2 : string
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    from sympy import symbols
    A = _parse_args(target=target, key=A1, default=0.0)
    B = _parse_args(target=target, key=A2, default=0.0)
    X = target[X]
    # Symbols used in symbolic function
    a, b, x = symbols('a,b,x')
    # Equation
    y = a*x + b
    # Callable functions
    r, s1, s2 = _build_func(eq=y, a=a, b=b,  x=x)
    # Values
    r_val = r(A, B, X)
    s1_val = s1(A, B, X)
    s2_val = s2(A, B, X)
    values = {'S1': s1_val, 'S2': s2_val, 'rate': r_val}
    return values


def power_law_sym(target, X, A1='', A2='', A3=''):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of *x*:

        .. math::
            r = A_{1} x^{A_{2}} + A_{3}

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    A1 -> A3 : string
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    from sympy import symbols
    A = _parse_args(target=target, key=A1, default=0.0)
    B = _parse_args(target=target, key=A2, default=1.0)
    C = _parse_args(target=target, key=A3, default=0.0)
    X = target[X]
    # Symbols used in symbolic function
    a, b, c, x = symbols('a,b,c,x')
    # Equation
    y = a*x**b + c
    # Callable functions
    r, s1, s2 = _build_func(eq=y, a=a, b=b, c=c, x=x)
    # Values
    r_val = r(A, B, C, X)
    s1_val = s1(A, B, C, X)
    s2_val = s2(A, B, C, X)
    values = {'S1': s1_val, 'S2': s2_val, 'rate': r_val}
    return values


def exponential_sym(target, X, A1='', A2='', A3='', A4='', A5='', A6=''):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of *x*:

        .. math::
            r = A_{1} A_{2}^{( A_{3} x^{ A_{4} } + A_{5})} + A_{6}

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    A1 -> A6 : string
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    from sympy import symbols
    A = _parse_args(target=target, key=A1, default=0.0)
    B = _parse_args(target=target, key=A2, default=1.0)
    C = _parse_args(target=target, key=A3, default=1.0)
    D = _parse_args(target=target, key=A4, default=1.0)
    E = _parse_args(target=target, key=A5, default=0.0)
    F = _parse_args(target=target, key=A6, default=0.0)
    X = target[X]
    # Symbols used in symbolic function
    a, b, c, d, e, f, x = symbols('a,b,c,d,e,f,x')
    # Equation
    y = a*b**(c*x**d + e) + f
    # Callable functions
    r, s1, s2 = _build_func(eq=y, a=a, b=b, c=c, d=d, e=e, f=f, x=x)
    # Values
    r_val = r(A, B, C, D, E, F, X)
    s1_val = s1(A, B, C, D, E, F, X)
    s2_val = s2(A, B, C, D, E, F, X)
    values = {'S1': s1_val, 'S2': s2_val, 'rate': r_val}
    return values


def natural_exponential_sym(target, X, A1='', A2='', A3='', A4='', A5=''):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of *x*:

        .. math::
            r = A_{1} exp( A_{2} x^{ A_{3} } + A_{4} )+ A_{5}

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    A1 -> A6 : string
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    from sympy import symbols, exp
    A = _parse_args(target=target, key=A1, default=0.0)
    B = _parse_args(target=target, key=A2, default=1.0)
    C = _parse_args(target=target, key=A3, default=1.0)
    D = _parse_args(target=target, key=A4, default=0.0)
    E = _parse_args(target=target, key=A5, default=0.0)
    X = target[X]
    # Symbols used in symbolic function
    a, b, c, d, e, x = symbols('a,b,c,d,e,x')
    # Equation
    y = a*exp(b*x**c + d) + e
    # Callable functions
    r, s1, s2 = _build_func(eq=y, a=a, b=b, c=c, d=d, e=e, x=x)
    # Values
    r_val = r(A, B, C, D, E, X)
    s1_val = s1(A, B, C, D, E, X)
    s2_val = s2(A, B, C, D, E, X)
    values = {'S1': s1_val, 'S2': s2_val, 'rate': r_val}
    return values


def logarithm_sym(target, X, A1='', A2='', A3='', A4='', A5='', A6=''):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of *x*:

        .. math::
            r = A_{1} Log_{ A_{2} }( A_{3} x^{ A_{4} }+ A_{5})+ A_{6}

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    A1 -> A6 : string
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    from sympy import symbols, log
    A = _parse_args(target=target, key=A1, default=0.0)
    B = _parse_args(target=target, key=A2, default=10.0)
    C = _parse_args(target=target, key=A3, default=1.0)
    D = _parse_args(target=target, key=A4, default=1.0)
    E = _parse_args(target=target, key=A5, default=0.0)
    F = _parse_args(target=target, key=A6, default=0.0)
    X = target[X]
    # Symbols used in symbolic function
    a, b, c, d, e, f, x = symbols('a,b,c,d,e,f,x')
    # Equation
    y = a * log((c * x**d + e), b) + f
    # Callable functions
    r, s1, s2 = _build_func(eq=y, a=a, b=b, c=c, d=d, e=e, f=f, x=x)
    # Values
    r_val = r(A, B, C, D, E, F, X)
    s1_val = s1(A, B, C, D, E, F, X)
    s2_val = s2(A, B, C, D, E, F, X)
    values = {'S1': s1_val, 'S2': s2_val, 'rate': r_val}
    return values


def natural_logarithm_sym(target, X, A1='', A2='', A3='', A4='', A5=''):
    r"""
    Calculates the rate, as well as slope and intercept of the following
    function at the given value of *x*:

        .. math::
            rate = A_{1} Ln( A_{2} x^{ A_{3} }+ A_{4})+ A_{5}

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    X : string
        The dictionary key on the target object containing the the quantity
        of interest

    A1 -> A5 : string
        The dictionary keys on the target object containing the coefficients
        values to be used in the source term model

    Returns
    -------
    A dictionary containing the following three items:

        **'rate'** - The value of the source term function at the given X.

        **'S1'** - The slope of the source term function at the given X.

        **'S2'** - The intercept of the source term function at the given X.

    The slope and intercept provide a linearized source term equation about the
    current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    """
    from sympy import symbols, ln
    A = _parse_args(target=target, key=A1, default=0.0)
    B = _parse_args(target=target, key=A2, default=1.0)
    C = _parse_args(target=target, key=A3, default=1.0)
    D = _parse_args(target=target, key=A4, default=0.0)
    E = _parse_args(target=target, key=A5, default=0.0)
    X = target[X]
    # Symbols used in symbolic function
    a, b, c, d, e, x = symbols('a,b,c,d,e,x')
    # Equation
    y = a * ln(b * x**c + d) + e
    # Callable functions
    r, s1, s2 = _build_func(eq=y, a=a, b=b, c=c, d=d, e=e, x=x)
    # Values
    r_val = r(A, B, C, D, E, X)
    s1_val = s1(A, B, C, D, E, X)
    s2_val = s2(A, B, C, D, E, X)
    values = {'S1': s1_val, 'S2': s2_val, 'rate': r_val}
    return values


def general_symbolic(target, eqn=None, arg_map=None):
    r'''
    A general function to interpret a sympy equation and evaluate the linear
    components of the source term.

    Parameters
    ----------
    target : OpenPNM object
        The OpenPNM object where the result will be applied.

    eqn : sympy symbolic expression for the source terms
        e.g. y = a*x**b + c

    arg_map : Dict mapping the symbols in the expression to OpenPNM data
        on the target. Must contain 'x' which is the independent variable.
        e.g. arg_map={'a':'pore.a', 'b':'pore.b', 'c':'pore.c', 'x':'pore.x'}

    Example
    ----------
    >>> import openpnm as op
    >>> from openpnm.models.physics import generic_source_term as gst
    >>> import numpy as np
    >>> import sympy
    >>> pn = op.network.Cubic(shape=[5, 5, 5], spacing=0.0001)
    >>> water = op.phases.Water(network=pn)
    >>> water['pore.a'] = 1
    >>> water['pore.b'] = 2
    >>> water['pore.c'] = 3
    >>> water['pore.x'] = np.random.random(water.Np)
    >>> a, b, c, x = sympy.symbols('a,b,c,x')
    >>> y = a*x**b + c
    >>> arg_map = {'a':'pore.a', 'b':'pore.b', 'c':'pore.c', 'x':'pore.x'}
    >>> water.add_model(propname='pore.general',
    ...                 model=gst.general_symbolic,
    ...                 eqn=y, arg_map=arg_map,
    ...                 regen_mode='normal')
    >>> assert 'pore.general.rate' in water.props()
    >>> assert 'pore.general.S1' in water.props()
    >>> assert 'pore.general.S1' in water.props()
    '''
    from sympy import postorder_traversal, srepr, symbols
    # First make sure all the symbols have been allocated dict items
    for arg in postorder_traversal(eqn):
        if srepr(arg)[:6] == 'Symbol':
            key = srepr(arg)[7:].strip('(').strip(')').strip("'")
            if key not in arg_map.keys():
                raise Exception('argument mapping incomplete, missing '+key)
    if 'x' not in arg_map.keys():
        raise Exception('argument mapping must contain "x" for the '
                        + 'independent variable')
    # Get the data
    data = {}
    args = {}
    for key in arg_map.keys():
        data[key] = target[arg_map[key]]
        # Callable functions
        args[key] = symbols(key)
    r, s1, s2 = _build_func(eqn, **args)
    r_val = r(*data.values())
    s1_val = s1(*data.values())
    s2_val = s2(*data.values())
    values = {'S1': s1_val, 'S2': s2_val, 'rate': r_val}
    return values


def butler_volmer_conc(
    target, X, z, j0, c_ref, alpha_anode, alpha_cathode,
    reaction_order=1,
    temperature="pore.temperature",
    reaction_area="pore.reaction_area",
    solid_voltage="pore.solid_voltage",
    electrolyte_voltage="pore.electrolyte_voltage",
    open_circuit_voltage="pore.open_circuit_voltage",
):
    r"""
    Calculates the rate, slope and intercept of the Butler-Volmer kinetic
    model based on **concentration** to be used in mass transfer
    algorithms.

        .. math::
            r = j_0 A_{rxn} (\frac{ X }{ c_{ref} }) ^ {\nu}
            \Big(
                \exp(  \frac{\alpha_a z F}{RT} \eta )
              - \exp( -\frac{\alpha_c z F}{RT} \eta )
            \Big)

    where:

        .. math::
            \eta = V_s - V_{\ell} - V_{oc}

    where ``V_s`` is the solid voltage, ``V_l`` is the electrolyte voltage,
    and ``V_oc`` is the open-circuit voltage.

    The slope and intercept provide a linearized source term equation
    about the current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    Parameters
    ----------
    target : GenericPhysics
        The Physics object where the result will be applied.
    X : str
        The dictionary key of the quantity of interest (i.e. main variable
        to be solved; in this case, concentration).
    z : float
        Number of electrons transferred in the redox reaction.
    j0 : float
        Exchange current density [A/m^2].
    c_ref : float
        Reference concentration [mol/m^3].
    alpha_anode : float
        Anodic transfer coefficient.
    alpha_cathode : float
        Cathodic transfer coefficient.
    reaction_order : float
        Reaction order, i.e. the exponent of the concentration term
    solid_voltage : str
        The dictionary key of solid phase voltages [V].
    electrolyte_voltage : str
        The dictionary key of electrolyte phase voltages [V].
    open_circuit_voltage : str
        The dictionary key of open-circuit voltage values [V].
    reaction_area : str
        The dictionary key of reaction area values [m^2].
    temperature : str
        The dictionary key of temperature values [K].

    Returns
    -------
    dict
        Dictionary containing the following key/value pairs:

        - rate : The value of the source term function at the given X.
        - S1 : The slope of the source term function at the given X.
        - S2 : The intercept of the source term function at the given X.

    Notes
    -----
    The difference between butler_volmer_conc and butler_volmer_voltage is
    that the former is linearized with respect to the electrolyte
    concentration whereas the latter is linearized with respect to the
    electrolyte voltage.

    Consequently, while the S1 and S2 produced by these model shouldn't be
    identical, they should both produce the same **rate** with the only
    difference that the rate generated by butler_volmer_conc has the units
    [mol/s] whereas that generated by butler_volmer_voltage has the units
    [C/s]. Therefore, the two rates will differ by z * F, where z is the
    number of electrons transferred and F is the Faraday's constant.

    """
    network = target.project.network
    pores = network.map_pores(pores=target.Ps, origin=target)

    # Fetch model variables
    X = target[X]
    T = target[temperature]
    Vs = target[solid_voltage]
    Ve = target[electrolyte_voltage]
    Voc = target[open_circuit_voltage]
    A_rxn = network[reaction_area][pores]
    F = _sp.constants.physical_constants["Faraday constant"][0]
    R = _sp.constants.R
    nu = reaction_order

    # Linearize with respect to X (electrolyte concentration)
    eta = Vs - Ve - Voc
    cte = j0 * A_rxn / (z * F)
    m1 = alpha_anode * z * F / (R * T)
    m2 = alpha_cathode * z * F / (R * T)
    fV = _np.exp(m1 * eta) - _np.exp(-m2 * eta)
    fC = (X / c_ref)**nu
    r = cte * fC * fV
    drdC = cte * (X / c_ref)**(nu - 1) * (1 / c_ref) * fV
    S1 = drdC
    S2 = r - drdC * X

    values = {"pore.S1": S1, "pore.S2": S2, "pore.rate": r}
    return values


def butler_volmer_voltage(
    target, X, z, j0, c_ref, alpha_anode, alpha_cathode,
    reaction_order=1,
    temperature="pore.temperature",
    reaction_area="pore.reaction_area",
    solid_voltage="pore.solid_voltage",
    open_circuit_voltage="pore.open_circuit_voltage",
    electrolyte_concentration="pore.electrolyte_concentration",
):
    r"""
    Calculates the rate, slope and intercept of the Butler-Volmer kinetic model
    based on **voltage** to be used in electron conduction algorithms.

        .. math::
            r = j_0 A_{rxn} (\frac{ c }{ c_{ref} }) ^ {\nu}
            \Big(
                \exp(  \frac{\alpha_a z F}{RT} \eta )
              - \exp( -\frac{\alpha_c z F}{RT} \eta )
            \Big)

    where:

        .. math::
            \eta = V_s - X - V_{oc}

    where ``V_s`` is the solid voltage, ``X`` is the electrolyte voltage,
    and ``V_oc`` is the open-circuit voltage.

    The slope and intercept provide a linearized source term equation
    about the current value of X as follow:

        .. math::
            rate = S_{1} X + S_{2}

    Parameters
    ----------
    target : GenericPhysics
        The Physics object where the result will be applied.
    X : str
        The dictionary key of the quantity of interest (i.e. main variable
        to be solved; in this case, electrolyte voltage).
    z : float
        Number of electrons transferred in the redox reaction.
    j0 : float
        Exchange current density [A/m^2].
    c_ref : float
        Reference concentration [mol/m^3].
    alpha_anode : float
        Anodic transfer coefficient.
    alpha_cathode : float
        Cathodic transfer coefficient.
    electrolyte_concentration : str
        The dictionary key of the electrolyte concentrations [mol/m^3].
    reaction_order : float
        Reaction order, i.e. the exponent of the concentration term
    solid_voltage : str
        The dictionary key of solid phase voltages [V].
    electrolyte_voltage : str
        The dictionary key of electrolyte phase voltages [V].
    open_circuit_voltage : str
        The dictionary key of open-circuit voltage values [V].
    reaction_area : str
        The dictionary key of reaction area values [m^2].
    temperature : str
        The dictionary key of temperature values [K].

    Returns
    -------
    dict
        Dictionary containing the following key/value pairs:

        - rate : The value of the source term function at the given X.
        - S1 : The slope of the source term function at the given X.
        - S2 : The intercept of the source term function at the given X.

    Notes
    -----
    The difference between butler_volmer_conc and butler_volmer_voltage is
    that the former is linearized with respect to the electrolyte
    concentration whereas the latter is linearized with respect to the
    electrolyte voltage.

    Consequently, while the S1 and S2 produced by these model shouldn't be
    identical, they should both produce the same **rate** with the only
    difference that the rate generated by butler_volmer_conc has the units
    [mol/s] whereas that generated by butler_volmer_voltage has the units
    [C/s]. Therefore, the two rates will differ by z * F, where z is the
    number of electrons transferred and F is the Faraday's constant.

    """
    network = target.project.network
    pores = network.map_pores(pores=target.Ps, origin=target)

    # Fetch model variables
    A_rxn = network[reaction_area][pores]
    Vs = target[solid_voltage]
    Voc = target[open_circuit_voltage]
    c = target[electrolyte_concentration]
    T = target[temperature]
    X = target[X]
    F = _sp.constants.physical_constants["Faraday constant"][0]
    R = _sp.constants.R
    nu = reaction_order

    # Linearize with respect to X (electrolyte voltage)
    eta = Vs - X - Voc
    cte = j0 * A_rxn
    m1 = alpha_anode * z * F / (R * T)
    m2 = alpha_cathode * z * F / (R * T)
    fV = _np.exp(m1 * eta) - _np.exp(-m2 * eta)
    dfVdV = -(m1 * _np.exp(m1 * eta) + m2 * _np.exp(-m2 * eta))
    fC = (c / c_ref)**nu
    r = cte * fC * fV
    drdV = cte * fC * dfVdV
    S1 = drdV
    S2 = r - drdV * X

    values = {"pore.S1": S1, "pore.S2": S2, "pore.rate": r}
    return values



r"""
Conduit shape factors' models to correct for the deviation from cylinders
in series assumption for Poisson-like physics such as mass diffusion or
electrical conduction, etc.

This submodule is deprecated and will be removed the near future. Use
`openpnm.models.geometry.diffusive_size_factors` instead.
"""
import numpy as _np

__all__ = ["ball_and_stick", "ball_and_stick_2d", "conical_frustum_and_stick"]


def ball_and_stick(
    target,
    pore_area="pore.area",
    throat_area="throat.area",
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    conduit_lengths="throat.conduit_lengths",
):
    r"""
    Calculates conduit shape factors for diffusive-like conductance.

    This model assumes pores and throats are spheres (balls) and constant
    cross-section cylinders (sticks). In addition to mass diffusion, this
    model can be used for other diffusion-like physics including thermal
    and electrical conductance.

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_area : str
        Dictionary key of the pore area values.
    throat_area : str
        Dictionary key of the throat area values.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.
    conduit_lengths : str
        Dictionary key of the conduit lengths' values.

    Returns
    -------
    dict
        Dictionary containing conduit shape factors. Shape factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    This model accounts for the variable cross-sectional area in spheres.

    This model could break if ``conduit_lengths`` does not correspond to
    an actual ball and stick. Example: pore length is greater than pore
    radius.

    """
    network = target.project.network
    throats = network.map_throats(throats=target.Ts, origin=target)
    cn = network["throat.conns"][throats]
    # Get pore diameter
    D1 = network[pore_diameter][cn[:, 0]]
    D2 = network[pore_diameter][cn[:, 1]]
    Dt = network[throat_diameter][throats]
    # Get conduit lengths
    L1 = network[conduit_lengths + ".pore1"][throats]
    L2 = network[conduit_lengths + ".pore2"][throats]
    Lt = network[conduit_lengths + ".throat"][throats]
    # Get pore/throat baseline areas (the one used in generic conductance)
    A1 = network[pore_area][cn[:, 0]]
    A2 = network[pore_area][cn[:, 1]]
    At = network[throat_area][throats]
    # Preallocating F, SF
    # F is INTEGRAL(1/A) dx , x : 0 --> L
    F1, F2, Ft = _np.zeros((3, len(Lt)))
    SF1, SF2, SFt = _np.ones((3, len(Lt)))
    # Setting SF to 1 when Li = 0 (ex. boundary pores)
    # INFO: This is needed since area could also be zero, which confuses NumPy
    m1, m2, mt = [Li != 0 for Li in [L1, L2, Lt]]
    SF1[~m1] = SF2[~m2] = SFt[~mt] = 1
    if (_np.sum(D1 <= 2 * L1) != 0) or (_np.sum(D2 <= 2 * L2) != 0):
        raise Exception(
            "Some pores can not be modeled with ball_and_stick"
            + " flow shape factor. Use another model for those pores"
            + " with (D/L)<=2"
        )
    # Handle the case where Dt >= Dp
    M1, M2 = [(Di <= Dt) & mi for Di, mi in zip([D1, D2], [m1, m2])]
    F1[M1] = (4 * L1 / (D1 * Dt * _np.pi))[M1]
    F2[M2] = (4 * L2 / (D2 * Dt * _np.pi))[M2]
    # Handle the rest (true balls and sticks)
    N1, N2 = [(Di > Dt) & mi for Di, mi in zip([D1, D2], [m1, m2])]
    F1[N1] = (2 / (D1 * _np.pi) * _np.arctanh(2 * L1 / D1))[N1]
    F2[N2] = (2 / (D2 * _np.pi) * _np.arctanh(2 * L2 / D2))[N2]
    Ft[mt] = (Lt / At)[mt]
    # Calculate conduit shape factors
    SF1[m1] = (L1 / (A1 * F1))[m1]
    SF2[m2] = (L2 / (A2 * F2))[m2]
    SFt[mt] = (Lt / (At * Ft))[mt]

    return {"pore1": SF1, "throat": SFt, "pore2": SF2}


def conical_frustum_and_stick(
    target,
    pore_area="pore.area",
    throat_area="throat.area",
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    conduit_lengths="throat.conduit_lengths",
):
    r"""
    Calculates conduit shape factors for diffusive-like conductance.

    This model assumes pores and throats are truncated pyramids (frustum)
    and constant cross-section cylinders (sticks). In addition to mass
    diffusion, this model can be used for other diffusion-like physics
    including thermal and electrical conductance.

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_area : str
        Dictionary key of the pore area values.
    throat_area : str
        Dictionary key of the throat area values.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.
    conduit_lengths : str
        Dictionary key of the conduit lengths' values.

    Returns
    -------
    dict
        Dictionary containing conduit shape factors. Shape factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    This model accounts for the variable cross-sectional area in pyramids.

    """
    network = target.project.network
    throats = network.map_throats(throats=target.Ts, origin=target)
    cn = network["throat.conns"][throats]
    # Get pore diameter
    D1 = network[pore_diameter][cn[:, 0]]
    D2 = network[pore_diameter][cn[:, 1]]
    Dt = network[throat_diameter][throats]
    # Get pore/throat baseline areas (the one used in generic conductance)
    A1 = network[pore_area][cn[:, 0]]
    A2 = network[pore_area][cn[:, 1]]
    At = network[throat_area][throats]
    # Get conduit lengths
    L1 = network[conduit_lengths + ".pore1"][throats]
    L2 = network[conduit_lengths + ".pore2"][throats]
    Lt = network[conduit_lengths + ".throat"][throats]
    # Preallocating F, SF
    # F is INTEGRAL(1/A) dx , x : 0 --> L
    F1, F2, Ft = _np.zeros((3, len(Lt)))
    SF1, SF2, SFt = _np.ones((3, len(Lt)))
    # Setting SF to 1 when Li = 0 (ex. boundary pores)
    # INFO: This is needed since area could also be zero, which confuses NumPy
    m1, m2, mt = [Li != 0 for Li in [L1, L2, Lt]]
    SF1[~m1] = SF2[~m2] = SFt[~mt] = 1
    # Handle the rest (non-zero-length conduits)
    F1[m1] = (4 * L1 / (D1 * Dt * _np.pi))[m1]
    F2[m2] = (4 * L2 / (D2 * Dt * _np.pi))[m2]
    Ft[mt] = (Lt / At)[mt]
    # Calculate conduit shape factors
    SF1[m1] = (L1 / (A1 * F1))[m1]
    SF2[m2] = (L2 / (A2 * F2))[m2]
    SFt[mt] = (Lt / (At * Ft))[mt]

    return {"pore1": SF1, "throat": SFt, "pore2": SF2}


def ball_and_stick_2d(
    target,
    pore_area="pore.area",
    throat_area="throat.area",
    pore_diameter="pore.diameter",
    throat_diameter="throat.diameter",
    conduit_lengths="throat.conduit_lengths",
):
    r"""
    Calculates conduit shape factors for diffusive-like conductance.

    This model assumes pores and throats are circles and rectangles. In
    addition to mass diffusion, this model can be used for other
    diffusion-like physics including thermal and electrical conductance.

    Parameters
    ----------
    target : GenericGeometry
        Geometry object which this model is associated with. This controls
        the length of the calculated array, and also provides access to
        other necessary properties.
    pore_area : str
        Dictionary key of the pore area values.
    throat_area : str
        Dictionary key of the throat area values.
    pore_diameter : str
        Dictionary key of the pore diameter values.
    throat_diameter : str
        Dictionary key of the throat diameter values.
    conduit_lengths : str
        Dictionary key of the conduit lengths' values.

    Returns
    -------
    dict
        Dictionary containing conduit shape factors. Shape factors are
        accessible via the keys: 'pore1', 'throat', and 'pore2'.

    Notes
    -----
    This model accounts for the variable cross-sectional area in circles.

    This model could break if ``conduit_lengths`` does not correspond to
    an actual ball and stick. Example: pore length is greater than pore
    radius.

    """
    network = target.project.network
    throats = network.map_throats(throats=target.Ts, origin=target)
    cn = network["throat.conns"][throats]
    # Get pore diameter
    D1 = network[pore_diameter][cn[:, 0]]
    D2 = network[pore_diameter][cn[:, 1]]
    # Get conduit lengths
    L1 = network[conduit_lengths + ".pore1"][throats]
    L2 = network[conduit_lengths + ".pore2"][throats]
    Lt = network[conduit_lengths + ".throat"][throats]
    # Get pore/throat baseline areas (the one used in generic conductance)
    A1 = network[pore_area][cn[:, 0]]
    A2 = network[pore_area][cn[:, 1]]
    At = network[throat_area][throats]
    # Preallocating F, SF
    # F is INTEGRAL(1/A) dx , x : 0 --> L
    F1, F2, Ft = _np.zeros((3, len(Lt)))
    SF1, SF2, SFt = _np.ones((3, len(Lt)))
    # Setting SF to 1 when Li = 0 (ex. boundary pores)
    # INFO: This is needed since area could also be zero, which confuses NumPy
    m1, m2, mt = [Li != 0 for Li in [L1, L2, Lt]]
    SF1[~m1] = SF2[~m2] = SFt[~mt] = 1
    F1[m1] = (0.5 * _np.arcsin(2 * L1 / D1))[m1]
    F2[m2] = (0.5 * _np.arcsin(2 * L2 / D2))[m2]
    Ft[mt] = (Lt / At)[mt]
    # Calculate conduit shape factors
    SF1[m1] = (L1 / (A1 * F1))[m1]
    SF2[m2] = (L2 / (A2 * F2))[m2]
    SFt[mt] = (Lt / (At * Ft))[mt]

    return {"pore1": SF1, "throat": SFt, "pore2": SF2}



r"""
Pore-scale models related to topology of the network.
"""
__all__ = ["coordination_number"]


def coordination_number(target):
    r"""
    """
    network = target.network
    N = network.num_neighbors(pores=network.Ps, flatten=False)
    vals = N[network.pores(target.name)]
    return vals



import openpnm
import numpy as np
from openpnm.utils import SettingsDict, logging
logger = logging.getLogger(__name__)


class SettingsDict(SettingsDict):
    def __setitem__(self, key, value):
        if key == 'loglevel':
            logger = logging.getLogger()
            logger.setLevel(value)
        super().__setitem__(key, value)


class Workspace(dict):
    r"""
    The Workspace object provides the highest level of adminstrative
    control over active OpenPNM sessions.

    It is a ``dictionary`` that stores a list of all open Projects by name.

    This class is a
    `singleton <https://en.wikipedia.org/wiki/Singleton_pattern>`_ so that
    whenever and wherever a Workspace is instantiated, the same instance is
    obtained.  This allows it to maintain a definitive record of all open
    Projects.

    See Also
    --------
    Project

    Notes
    -----
    The Workspace object contains a variety of functions that one might expect
    from the 'file-menu' in a typical GUI.

    """

    __instance__ = None

    def __new__(cls, *args, **kwargs):
        if Workspace.__instance__ is None:
            Workspace.__instance__ = dict.__new__(cls)
        return Workspace.__instance__

    def __init__(self):
        super().__init__()
        self._projects = {}
        self.settings = SettingsDict()
        self.settings['loglevel'] = 30

    def __setitem__(self, name, project):
        if name is None:
            name = self._gen_name()
        if name in self.keys():
            # To ensure renaming a Project to its current name is allowed
            if self.__instance__[name] is not project:
                raise Exception(f"A project named {name} already exists")
        if project in self.values():
            self.pop(project.name, None)
        if not isinstance(project, openpnm.utils.Project):
            project = openpnm.utils.Project(project, name=name)
        super().__setitem__(name, project)

    def copy(self):
        r"""
        """
        raise Exception('Cannot copy Workspace, only one can exist at a time')

    def _create_console_handles(self, project):
        r"""
        Adds all objects in the given project to the console as variables
        with handle names taken from each object's name.
        """
        import __main__
        for item in project:
            __main__.__dict__[item.name] = item

    @property
    def version(self):
        return openpnm.__version__

    def save_workspace(self, filename=None):
        r"""
        Save all projects in the current workspace as a single file

        Parameters
        ----------
        filename : str
            The filename to use when saving.  If not provided, the present
            date and time are used.

        Notes
        -----
        The file is actually zip archive containing ``pnm`` files, one for
        each project in the workspace. This archive can be extracted and each
        ``pnm`` file can be loaded manually using ``load_project`` or the
        ``openpnm.io.PNM`` class.
        """
        from datetime import datetime
        if filename is None:
            dt = datetime.now()
            filename = dt.strftime("%Y_%m_%d_%H_%M_%S")
        from zipfile import ZipFile
        with ZipFile(filename + '.wrk', 'w') as z:
            for prj in self.values():
                prj.save_project()
                z.write(prj.name + '.pnm')

    def load_workspace(self, filename):
        r"""
        Load project(s) from a saved workspace into current workspace

        Parameters
        ----------
        filename : str or path object
            The filename containing the saved workspace

        Notes
        -----
        ??

        """
        from zipfile import ZipFile
        with ZipFile(filename, 'r') as z:
            logger.info('Loading projects contained in ' + filename)
            files = z.filelist
            for f in files:
                self.load_project(f.orig_filename)

    def save_project(self, project, filename=None):
        r"""
        Saves given Project to a ``pnm`` file

        This will include all of associated objects, including algorithms.

        Parameters
        ----------
        project : OpenPNM Project
            The project to save.

        filename : string, optional
            If no filename is given, the given project name is used. See Notes
            for more information.

        See Also
        --------
        save_workspace

        Notes
        -----
        The filename can be a string such as 'saved_file.pnm'.  The string can
        include absolute path such as 'C:\networks\saved_file.pnm', or can
        be a relative path such as '..\..\saved_file.pnm', which will look
        2 directories above the current working directory.  Can also be a
        path object object such as that produced by ``pathlib`` or
        ``os.path`` in the Python standard library.

        """
        from openpnm.io import PNM
        PNM.save_project(project=project, filename=filename)

    def load_project(self, filename, overwrite=False):
        r"""
        Loads a Project from the specified 'pnm' file

        The loaded project is added to the Workspace . This will *not* delete
        any existing Projects in the Workspace and will rename any Projects
        being loaded if necessary.

        Parameters
        ----------
        filename : string or path object
            The name of the file to open.  See Notes for more information.

        See Also
        --------
        load_workspace

        Notes
        -----
        The filename can be a string such as 'saved_file.pnm'.  The string can
        include absolute path such as 'C:\networks\saved_file.pnm', or can
        be a relative path such as '..\..\saved_file.pnm', which will look
        2 directories above the current working directory.  Can also be a
        path object object such as that produced by ``pathlib`` or
        ``os.path`` in the Python standard library.

        """
        from openpnm.io import PNM
        proj = PNM.load_project(filename=filename)
        return proj

    def close_project(self, project):
        r"""
        Removes the specified Project from the Workspace

        This does not save the project, so any changes will be lost.
        """
        del self[project.name]

    def copy_project(self, project, name=None):
        r"""
        Make a copy of an existing Project

        Parameters
        ----------
        project : Project object
            The Project object to be copied

        name : string, optional
            A name for the new copy of the project.  If not supplied, then
            one will be generated (e.g. 'proj_02')

        Returns
        -------
        proj : list
            A handle to the new Project

        """
        proj = project.copy(name)
        return proj

    def new_project(self, name=None):
        r"""
        Creates a new empty Project object

        Parameters
        ----------
        name : string (optional)
            The unique name to give to the project.  If none is given, one
            will be automatically generated (e.g. 'proj_01`)

        Returns
        -------
        proj : list
            An empty Project object, suitable for passing into a Network
            generator

        """
        proj = openpnm.utils.Project(name=name)
        return proj

    def _gen_name(self):
        r"""
        Generates a valid name for projects
        """
        n = [0]
        for item in self.keys():
            if item.startswith('proj_'):
                n.append(int(item.split('proj_')[1]))
        name = 'proj_'+str(max(n)+1).zfill(2)
        return name

    def _gen_ids(self, size):
        r"""
        Generates a sequence of integers of the given ``size``, starting at 1
        greater than the last produced value.

        The Workspace object keeps track of the most recent value, which
        persists until the current python session is restarted, so the
        returned array contains unique values for the given session.

        Parameters
        ----------
        size : int
            The number of values to generate.

        Returns
        -------
        A Numpy array of the specified size, containing integer values starting
        from the last used values.

        Notes
        -----
        When a new Workspace is created the
        """
        if not hasattr(self, '_next_id'):
            # If _next_id has not been set, then assign it
            self._next_id = 0
            # But check ids in any objects present first
            for proj in self.values():
                if len(proj) > 0:
                    if 'pore._id' in proj.network.keys():
                        Pmax = proj.network['pore._id'].max() + 1
                        Tmax = proj.network['throat._id'].max() + 1
                        self._next_id = max([Pmax, Tmax, self._next_id])
        ids = np.arange(self._next_id, self._next_id + size, dtype=np.int64)
        self._next_id += size
        return ids

    def __str__(self):
        s = []
        hr = '―'*78
        s.append(hr)
        s.append('OpenPNM Version ' + openpnm.__version__ + ' Workspace')
        s.append(hr)
        for item in self.values():
            s.append(' ' + item.name)
            s.append(item.__str__())
        return '\n'.join(s)



from openpnm.core import Subdomain, ModelsMixin
from openpnm.utils import Workspace, logging
logger = logging.getLogger(__name__)
ws = Workspace()


class GenericGeometry(Subdomain, ModelsMixin):
    r"""
    This generic class is meant as a starter for custom Geometry objects

    It has no pore-scale models assigned to it, so a a blank slate.  Note that
    all OpenPNM Geometry sub-classes are just GenericGeometry instances with a
    number of models added.

    Parameters
    ----------
    network : GenericNetwork
        The Network object to which this Geometry applies.
    pores : array_like
        The list of pores where this Geometry applies.
    throats : array_like
        The list of throats where this Geometry applies.
    name : str
        A unique name to apply to the object.  This name will also be used as a
        label to identify where this Geometry applies.
    project : Project, optional
        A Project can be specified instead of ``network``.

    Examples
    --------
    >>> import openpnm as op
    >>> pn = op.network.Cubic(shape=[5, 5, 5])
    >>> Ps = pn.pores('all')    # Get all pores
    >>> Ts = pn.throats('all')  # Get all throats
    >>> geom = op.geometry.GenericGeometry(network=pn, pores=Ps, throats=Ts)

    Now assign pore-scale models to the empty object:

    >>> geom.add_model(propname='pore.size',
    ...                model=op.models.misc.random,
    ...                element='pore',
    ...                num_range=[0.01, 0.1])

    Confirm that the object has one added model:

    >>> print(geom.models)
    ―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    #   Property Name                       Parameter                 Value
    ―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    1   pore.size                           model:                    random
                                            element:                  pore
                                            num_range:                [0.01, 0.1]
                                            seed:                     None
                                            regeneration mode:        normal
    ―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――

    The results of the model can be seen using the ``show_hist`` function:

    >>> import matplotlib as mpl
    >>> mpl.use('Agg')
    >>> geom.show_hist('pore.size')

    .. image:: /../docs/static/images/generic_geometry_histogram.png
        :width: 500px
        :align: center

    """

    def __init__(self, network=None, project=None, pores=[], throats=[],
                 settings={}, **kwargs):
        # Define some default settings
        self.settings.update({'prefix': 'geo'})
        # Overwrite with user supplied settings, if any
        self.settings.update(settings)

        # Deal with network or project arguments
        if network is not None:
            if project is not None:
                assert network is project.network
            else:
                project = network.project

        super().__init__(project=project, **kwargs)

        network = self.project.network
        if network:
            network[f'pore.{self.name}'] = False
            network[f'throat.{self.name}'] = False
            try:
                self.add_locations(pores=pores, throats=throats)
            except Exception as e:
                network.project.purge_object(self)
                logger.error(f'{e}, instantiation cancelled')

    def add_locations(self, pores=[], throats=[]):
        r"""
        Adds associations between this geometry and the given pore and/or
        throat locations.

        Parameters
        ----------
        pores and throats : array_like
            The pore and/or throat locations for which the association should
            be added.  These indices are for the full domain.

        Notes
        -----
        If a physics object is associated with this geometry, then its
        pore and/or throat associations are also changed.
        """
        pores = self.network._parse_indices(pores)
        throats = self.network._parse_indices(throats)
        objects = self.project.find_physics(self)
        objects.append(self)
        for obj in objects:
            if len(pores) > 0:
                obj._set_locations(element='pore', indices=pores, mode='add')
            if len(throats) > 0:
                obj._set_locations(element='throat', indices=throats, mode='add')

    def drop_locations(self, pores=[], throats=[]):
        r"""
        Removes association between this geometry and the given pore and/or
        throat locations.

        Parameters
        ----------
        pores and throats : array_like
            The pore and/or throat locations from which the association should
            be removed.  These indices refer to the full domain.

        Notes
        -----
        If a physics object is associated with this geometry, then its
        pore and/or throat associations are also changed.

        """
        pores = self.network._parse_indices(pores)
        throats = self.network._parse_indices(throats)
        objects = self.project.find_physics(self)
        objects.append(self)
        for obj in objects:
            if len(pores) > 0:
                obj._set_locations(element='pore', indices=pores, mode='drop')
            if len(throats) > 0:
                obj._set_locations(element='throat', indices=throats, mode='drop')



import numpy as np
from openpnm.algorithms import GenericTransport
# Uncomment this line when we stop supporting Python 3.6
# from dataclasses import dataclass, field
# from typing import List
from openpnm.utils import logging, Docorator, GenericSettings
docstr = Docorator()
logger = logging.getLogger(__name__)


# class RelaxationSettings(GenericSettings):
#     r"""
#     This class is a demonstration of how we can add nested settings classes
#     to other settings classes to make categories for some settings.  This is
#     being appended to the ReactiveTransportSettings class under the
#     'relaxation' attribute, and it works as planned by allowing the nested
#     dot access to its parameters. More work would be required to get it
#     functional such as dealing with deeply nested dicts and so on, but it
#     works in principal.
#     """
#     source = 1.0
#     quantity = 1.0


@docstr.get_sections(base='ReactiveTransportSettings',
                     sections=['Parameters', 'Other Parameters'])
@docstr.dedent
# Uncomment this line when we stop supporting Python 3.6
# @dataclass
class ReactiveTransportSettings(GenericSettings):
    r"""

    Parameters
    ----------
    %(GenericTransportSettings.parameters)s

    quantity : str
        The name of the physical quantity to be calculated
    conductance : str
        The name of the pore-scale transport conductance values. These are
        typically calculated by a model attached to a *Physics* object
        associated with the given *Phase*.

    Other Parameters
    ----------------
    sources : list
        List of source terms that have been added
    relaxation_source : float (default = 1.0)
        A relaxation factor to control under-relaxation of the source term.
        Factor approaching 0 leads to improved stability but slower simulation.
        Factor approaching 1 gives fast simulation but may be unstable.
    relaxation_quantity : float (default = 1.0)
        A relaxation factor to control under-relaxation for the quantity
        solving for. Factor approaching 0 leads to improved stability but
        slower simulation. Factor approaching 1 gives fast simulation but
        may be unstable.
    nlin_max_iter : int
        Maximum number of iterations allowed for the nonlinear solver to
        converge. This parameter is different that ``GenericTransport``'s
        ``solver_max_iter``.

    ----

    **The following parameters pertain to the ``GenericTransport`` class**

    %(GenericTransportSettings.other_parameters)s

    """

    nlin_max_iter = 5000
    # relaxation = RelaxationSettings()
    relaxation_source = 1.0
    relaxation_quantity = 1.0
    # Swap the following 2 lines when we stop supporting Python 3.6
    # sources: List = field(default_factory=lambda: [])
    sources = []


@docstr.get_sections(base='ReactiveTransport', sections=['Parameters'])
@docstr.dedent
class ReactiveTransport(GenericTransport):
    r"""
    A subclass for steady-state simulations with (optional) source terms.

    Parameters
    ----------
    %(GenericTransport.parameters)s

    Notes
    -----
    This subclass performs steady simulations of transport phenomena with
    reactions when source terms are added.

    """

    def __init__(self, settings={}, phase=None, **kwargs):
        super().__init__(**kwargs)
        self.settings._update_settings_and_docs(ReactiveTransportSettings)
        self.settings.update(settings)
        if phase is not None:
            self.setup(phase=phase)

    @docstr.get_sections(base='ReactiveTransport.setup',
                         sections=['Parameters', 'Notes'])
    @docstr.dedent
    def setup(self, phase=None, quantity='', conductance='',
              nlin_max_iter=None, relaxation_source=None,
              relaxation_quantity=None, **kwargs):
        r"""
        This method takes several arguments that are essential to running
        the algorithm and adds them to the settings.

        Parameters
        ----------
        %(GenericTransportSettings.parameters)s
        %(ReactiveTransportSettings.parameters)s

        Notes
        -----
        Under-relaxation is a technique used for improving stability of a
        computation, particularly in the presence of highly non-linear
        terms. Under-relaxation used here limits the change in a variable
        from one iteration to the next. An optimum choice of the
        relaxation factor is one that is small enough to ensure stable
        simulation and large enough to speed up the computation.

        """
        if phase:
            self.settings['phase'] = phase.name
        if quantity:
            self.settings['quantity'] = quantity
        if conductance:
            self.settings['conductance'] = conductance
        if nlin_max_iter:
            self.settings['nlin_max_iter'] = nlin_max_iter
        if relaxation_source:
            self.settings['relaxation_source'] = relaxation_source
        if relaxation_quantity:
            self.settings['relaxation_quantity'] = relaxation_quantity
        super().setup(**kwargs)

    def run(self, x0=None):
        r"""
        Builds the A and b matrices, and calls the solver specified in the
        ``settings`` attribute.

        Parameters
        ----------
        x0 : ND-array
            Initial guess of unknown variable
        """
        self._validate_settings()
        # Check if A and b are well-defined
        self._validate_data_health()
        quantity = self.settings['quantity']
        logger.info('Running ReactiveTransport')
        x0 = np.zeros(self.Np, dtype=float) if x0 is None else x0
        self["pore.initial_guess"] = x0
        x = self._run_reactive(x0)
        self[quantity] = x

    @docstr.dedent
    def reset(self, source_terms=False, **kwargs):
        r"""
        %(GenericTransport.reset.full_desc)s

        Parameters
        ----------
        %(GenericTransport.reset.parameters)s
        source_terms : boolean
            If ``True`` removes source terms. The default is ``False``.

        """
        super().reset(**kwargs)
        if source_terms:
            # Remove item from label dictionary
            for item in self.settings['sources']:
                self.pop(item)
            # Reset the settings dict
            self.settings['sources'] = []

    def set_source(self, propname, pores, mode='overwrite'):
        r"""
        Applies a given source term to the specified pores.

        Parameters
        ----------
        propname : string
            The property name of the source term model to be applied.
        pores : array_like
            The pore indices where the source term should be applied.
        mode : str
            Controls how the sources are applied. Options are:

            'merge' - Adds supplied source term to already existing ones.
            'overwrite' - (default) Deletes all existing source terms of the
            given ``propname`` then adds the specified new ones

        Notes
        -----
        Source terms cannot be applied in pores where boundary conditions have
        already been set. Attempting to do so will result in an error being
        raised.

        """
        locs = self.tomask(pores=pores)
        # Check if any BC is already set in the same locations
        locs_BC = np.isfinite(self['pore.bc_value']) + np.isfinite(self['pore.bc_rate'])
        if (locs & locs_BC).any():
            raise Exception('Boundary conditions already present in given '
                            + 'pores, cannot also assign source terms')
        if propname not in self.keys():
            self[propname] = False
        if mode == 'merge':
            self[propname][locs] = True
        elif mode == 'overwrite':
            self[propname] = locs
        # Check if propname already in source term list
        if propname not in self.settings['sources']:
            self.settings['sources'].append(propname)

    def remove_source(self, propname, pores=None):
        r"""
        Removes source terms from specified pores.

        Parameters
        ----------
        propname : str
            The property name of the source term model to be removed.
        pores : array_like
            The pore indices where the source term should be applied.

        """
        if pores is None:
            pores = self.Ps
        locs = self.tomask(pores=pores)
        if propname not in self.keys():
            self[propname] = False
        self[propname][locs] = False

    def _update_iterative_props(self):
        """r
        Update physics using the current value of ``quantity``.

        Notes
        -----
        The algorithm directly writes the value of 'quantity' into the
        phase. This method was implemented relaxing one of the OpenPNM
        rules of algorithms not being able to write into phases.

        """
        phase = self.project.phases()[self.settings['phase']]
        physics = self.project.find_physics(phase=phase)
        geometries = self.project.geometries().values()
        # Regenerate iterative props with new guess
        iterative_props = self._get_iterative_props()
        if len(iterative_props) > 0:
            # Put quantity on phase so physics finds it when regenerating
            key = self.settings['quantity']
            phase[key] = self[key]
            phase.regenerate_models(propnames=iterative_props)
            for geometry in geometries:
                geometry.regenerate_models(iterative_props)
            for phys in physics:
                phys.regenerate_models(iterative_props)

    def _apply_sources(self):
        """r
        Update ``A`` and ``b`` applying source terms to specified pores.

        Notes
        -----
        Applying source terms to ``A`` and ``b`` is performed after
        (optionally) under-relaxing the source term to improve numerical
        stability. Physics are also updated before applying source terms
        to ensure that source terms values are associated with the current
        value of 'quantity'.

        For source term under-relaxation, old values of S1 and S2 need
        to be stored somewhere, we chose to store them on the algorithm
        object. This is because storing them on phase/physics creates
        unintended problems, ex. storing them on physics -> IO complains
        added depth to the NestedDict, and storing them on the phase
        object results in NaNs in case source term is only added to a
        subset of nodes, which breaks our _check_for_nans algorithm.

        Warnings
        --------
        In the case of a transient simulation, the updates in ``A`` and ``b``
        also depend on the time scheme. So, ``_correct_apply_sources()`` needs to
        be run afterwards to correct the already applied relaxed source terms.

        """
        phase = self.project.phases()[self.settings['phase']]
        w = self.settings['relaxation_source']

        for item in self.settings['sources']:
            element, prop = item.split(".")
            _item = ".".join([element, "_" + prop])
            first_iter = False if _item + ".S1.old" in self.keys() else True
            Ps = self.pores(item)
            # Fetch S1/S2 and their old values (don't exist on 1st iter)
            S1 = phase[item + ".S1"][Ps]
            S2 = phase[item + ".S2"][Ps]
            X1 = self[_item + ".S1.old"][Ps] if not first_iter else S1
            X2 = self[_item + ".S2.old"][Ps] if not first_iter else S2
            # Source term relaxation
            S1 = phase[item + '.S1'][Ps] = w * S1 + (1.0 - w) * X1
            S2 = phase[item + '.S2'][Ps] = w * S2 + (1.0 - w) * X2
            # Modify A and b based on "relaxed" S1/S2
            datadiag = self._A.diagonal().copy()
            datadiag[Ps] = datadiag[Ps] - S1
            self._A.setdiag(datadiag)
            self._b[Ps] = self._b[Ps] + S2
            # Replace old values of S1/S2 by their current values
            self[_item + ".S1.old"] = phase[item + ".S1"]
            self[_item + ".S2.old"] = phase[item + ".S2"]

    def _run_reactive(self, x0):
        r"""
        Repeatedly updates ``A``, ``b``, and the solution guess within
        according to the applied source term then calls ``_solve`` to
        solve the resulting system of linear equations.

        Stops when the residual falls below ``solver_tol * norm(b)`` or
        when the maximum number of iterations is reached.

        Parameters
        ----------
        x0 : ndarray
            Initial guess of unknown variable

        Returns
        -------
        x : ndarray
            Solution array.

        Notes
        -----
        The algorithm must at least complete one iteration, and hence the
        check for itr >= 1, because otherwise, _check_for_nans() never
        gets called in case there's something wrong with the data, and
        therefore, the user won't get notified about the root cause of the
        algorithm divergence.

        """
        w = self.settings['relaxation_quantity']
        quantity = self.settings['quantity']
        max_it = self.settings['nlin_max_iter']
        # Write initial guess to algorithm obj (for _update_iterative_props to work)
        self[quantity] = x = x0
        # Update A and b based on self[quantity]
        self._update_A_and_b()
        # Just in case you got a lucky guess, i.e. x0!
        if self._is_converged():
            logger.info(f'Solution converged: {self._get_residual():.4e}')
            return x

        for itr in range(max_it):
            # Solve, use relaxation, and update solution on algorithm obj
            self[quantity] = x = self._solve(x0=x) * w + x * (1 - w)
            self._update_A_and_b()
            # Check solution convergence
            if self._is_converged():
                logger.info(f'Solution converged: {self._get_residual():.4e}')
                return x
            logger.info(f'Tolerance not met: {self._get_residual():.4e}')

        if not self._is_converged():
            raise Exception(f"Not converged after {max_it} iterations.")

    def _update_A_and_b(self):
        r"""
        Updates A and b based on the most recent solution stored on
        algorithm object.
        """
        # Update iterative properties on phase, geometries, and physics
        self._update_iterative_props()
        # Build A and b, apply BCs/source terms
        self._build_A()
        self._build_b()
        self._apply_BCs()
        self._apply_sources()

    def _get_iterative_props(self):
        r"""
        Find and return properties that need to be iterated while running
        the algorithm.

        Notes
        -----
        This method was moved from ReactiveTransport class to
        GenericTransport because source terms are not necessarily the only
        properties that need iteration during an algorithm (ex.
        concentration-dependent conductance)

        """
        import networkx as nx
        phase = self.project.phases(self.settings['phase'])
        physics = self.project.find_physics(phase=phase)
        geometries = self.project.geometries().values()
        # Combine dependency graphs of phase and all physics/geometries
        dg = phase.models.dependency_graph(deep=True)
        for g in geometries:
            dg = nx.compose(dg, g.models.dependency_graph(deep=True))
        for p in physics:
            dg = nx.compose(dg, p.models.dependency_graph(deep=True))
        base_props = [self.settings["quantity"]]
        if base_props is None:
            return []
        # Find all props downstream that rely on "quantity"
        dg = nx.DiGraph(nx.edge_dfs(dg, source=base_props))
        if len(dg.nodes) == 0:
            return []
        iterative_props = list(nx.dag.lexicographical_topological_sort(dg))
        # "quantity" shouldn't be in the returned list but "variable_props" should
        try:
            iterative_props.remove(self.settings["quantity"])
        except ValueError:
            pass
        return iterative_props

    @docstr.dedent
    def _set_BC(self, pores, bctype, bcvalues=None, mode='merge'):
        r"""
        Apply boundary conditions to specified pores if no source terms
        are already assigned to these pores. Otherwise, raise an error.

        Parameters
        ----------
        %(GenericTransport._set_BC.parameters)s

        Notes
        -----
        %(GenericTransport._set_BC.notes)s

        """
        # First check that given pores do not have source terms already set
        for item in self.settings['sources']:
            if np.any(self[item][pores]):
                raise Exception(
                    'Source term already present in given pores, cannot also'
                    ' assign boundary conditions'
                )
        # Then call parent class function if above check passes
        super()._set_BC(pores=pores, bctype=bctype, bcvalues=bcvalues, mode=mode)



import numpy as np
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import logging, Docorator, GenericSettings
docstr = Docorator()
logger = logging.getLogger(__name__)


@docstr.get_sections(base='NernstPlanckSettings', sections=['Parameters'])
@docstr.dedent
class NernstPlanckSettings(GenericSettings):
    r"""
    Parameters
    ----------
    %(ReactiveTransportSettings.parameters)s
    quantity : str, optional
        The quantity to solve for. The default value is
        'pore.concentration'. Note that this will have the 'ion' name
        appended to the end (i.e. ``pore.concentration.Na``)
    conductance : str, optional
        The overall mass transfer conductance of the ion. The default
        value is 'throat.ad_dif_mig_conductance'.
    diffusive_conductance : str
        The name of the diffusive conductance values. The default value is
        'throat.diffusive_conductance'.
    hydraulic_conductance : str, optional
        The name of the hydraulic conductance values. The default value is
        'throat.hydraulic_conductance'.
    pressure : str, optional
        The name of the pressure values calculated by the ``StokesFlow``
        algorithm. The default value is 'pore.pressure'.

    Other Parameters
    ----------------
    s_scheme : str
        The discretization scheme for the advective terms. The default
        value is 'powerlaw'.

    ----

    **The following parameters pertain to the ReactiveTransport class**

    %(ReactiveTransportSettings.other_parameters)s

    ----

    **The following parameters pertain to the GenericTransport class**

    %(GenericTransportSettings.other_parameters)s

    """
    ion = ''
    quantity = 'pore.concentration'
    conductance = 'throat.ad_dif_mig_conductance'
    hydraulic_conductance = 'throat.hydraulic_conductance'
    diffusive_conductance = 'throat.diffusive_conductance'
    pressure = 'pore.pressure'


class NernstPlanck(ReactiveTransport):
    r"""
    A subclass of ``ReactiveTransport`` to simulate transport of charged
    species (such as ions) in dilute solutions.

    """

    def __init__(self, ion=None, settings={}, **kwargs):
        super().__init__(**kwargs)
        # self.name = electrolyte  # This interfers with component name
        self.settings._update_settings_and_docs(NernstPlanckSettings())
        self.settings.update(settings)
        # Parse the given ion and append name to quantity and conductance
        if ion:
            if not isinstance(ion, str):  # Convert ion object to str
                ion = ion.name
            self.settings['ion'] = ion
        quantity = self.settings['quantity']
        if (ion is not None and not quantity.endswith(ion)):
            quantity = '.'.join(quantity.split('.')[:2])
            quantity += ('.' + ion)  # Re-add ion name
            self.settings['quantity'] = quantity  # Add full value to settings
        conductance = self.settings['conductance']
        if (ion is not None and not conductance.endswith(ion)):
            conductance = '.'.join(conductance.split('.')[:2])
            conductance += ('.' + ion)  # Re-add ion name
            self.settings['conductance'] = conductance
        diffusive_conductance = self.settings['diffusive_conductance']
        if (ion is not None and not diffusive_conductance.endswith(ion)):
            diffusive_conductance = '.'.join(
                diffusive_conductance.split('.')[:2])
            diffusive_conductance += ('.' + ion)  # Re-add ion name
            self.settings['diffusive_conductance'] = diffusive_conductance

    def setup(self, phase=None, quantity='', conductance='',
              diffusive_conductance='', ion='', **kwargs):
        r"""

        Parameters
        ----------
        %(NernstPlanckSettings.parameters)s

        """
        if phase:
            self.settings['phase'] = phase.name
        if quantity:
            self.settings['quantity'] = quantity  # Add full value to settings
        if conductance:
            self.settings['conductance'] = conductance
        if diffusive_conductance:
            self.settings['diffusive_conductance'] = diffusive_conductance
        if ion:
            self.settings['quantity'] = quantity
        super().setup(**kwargs)

    def set_outflow_BC(self, pores, mode='merge'):
        r"""
        Adds outflow boundary condition to the selected pores.

        Notes
        -----
        Outflow condition simply means that the gradient of the solved
        quantity does not change, i.e. is 0.

        """
        # Hijack the parse_mode function to verify mode/pores argument
        allowed_modes = ['merge', 'overwrite', 'remove']
        mode = self._parse_mode(mode, allowed=allowed_modes, single=True)
        pores = self._parse_indices(pores)

        # Calculating A[i,i] values to ensure the outflow condition
        network = self.project.network
        phase = self.project.phases()[self.settings['phase']]
        throats = network.find_neighbor_throats(pores=pores)
        C12 = network.conns[throats]
        P12 = phase[self.settings['pressure']][C12]
        # Note: only keep the advection term when flow leaves the outflow pores
        gh = phase[self.settings['hydraulic_conductance']][throats]
        Q12 = -gh * np.diff(P12, axis=1).squeeze()
        Qp = np.zeros(self.Np)
        np.add.at(Qp, C12[:, 0], -Q12)
        np.add.at(Qp, C12[:, 1], Q12)

        # Store boundary values
        if ('pore.bc_outflow' not in self.keys()) or (mode == 'overwrite'):
            self['pore.bc_outflow'] = np.nan
        self['pore.bc_outflow'][pores] = Qp[pores]

    def _apply_BCs(self):
        # Apply Dirichlet and rate BCs
        ReactiveTransport._apply_BCs(self)
        if 'pore.bc_outflow' not in self.keys():
            return
        # Apply outflow BC
        diag = self.A.diagonal()
        ind = np.isfinite(self['pore.bc_outflow'])
        diag[ind] += self['pore.bc_outflow'][ind]
        self.A.setdiag(diag)



r"""
"""
import numpy as np
from openpnm.utils import logging
logger = logging.getLogger(__name__)


def from_neighbor_throats(target, prop, mode='min', ignore_nans=True):
    r"""
    Adopt a value from the values found in neighboring throats

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    prop : string
        The dictionary key of the array containing the throat property to be
        used in the calculation.
    mode : string
        Controls how the pore property is calculated.  Options are 'min',
        'max' and 'mean'.

    Returns
    -------
    value : ND-array
        Array containing customized values based on those of adjacent throats.

    """
    prj = target.project
    network = prj.network
    boss = prj.find_full_domain(target)
    data = boss[prop]
    nans = np.isnan(data)
    im = network.create_incidence_matrix()
    if mode == 'min':
        if ignore_nans:
            data[nans] = np.inf
        values = np.ones((network.Np, ))*np.inf
        np.minimum.at(values, im.row, data[im.col])
    if mode == 'max':
        if ignore_nans:
            data[nans] = -np.inf
        values = np.ones((network.Np, ))*-np.inf
        np.maximum.at(values, im.row, data[im.col])
    if mode == 'mean':
        if ignore_nans:
            data[nans] = 0
        values = np.zeros((network.Np, ))
        np.add.at(values, im.row, data[im.col])
        counts = np.zeros((network.Np, ))
        np.add.at(counts, im.row, np.ones((network.Nt, ))[im.col])
        if ignore_nans:
            np.subtract.at(counts, im.row, nans[im.col])
        values = values/counts
    Ps = boss.map_pores(target.pores(), target)
    return np.array(values)[Ps]


def from_neighbor_pores(target, prop, mode='min', ignore_nans=True):
    r"""
    Adopt a value based on the values in neighboring pores

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    prop : string
        The dictionary key to the array containing the pore property to be
        used in the calculation.
    mode : string
        Controls how the throat property is calculated.  Options are 'min',
        'max' and 'mean'.
    ignore_nans : boolean (default is ``True``)
        If ``True`` the result will ignore ``nans`` in the neighbors

    Returns
    -------
    value : ND-array
        Array containing customized values based on those of adjacent pores.

    """
    prj = target.project
    lookup = prj.find_full_domain(target)
    network = prj.network
    throats = network.map_throats(target.throats(), target)
    P12 = network.find_connected_pores(throats)
    pvalues = lookup[prop][P12]
    if ignore_nans:
        pvalues = np.ma.MaskedArray(data=pvalues, mask=np.isnan(pvalues))
    try:  # If pvalues is not empty
        if mode == 'min':
            value = np.amin(pvalues, axis=1)
        if mode == 'max':
            value = np.amax(pvalues, axis=1)
        if mode == 'mean':
            value = np.mean(pvalues, axis=1)
    except np.AxisError:  # Handle case of empty pvalues
        value = []
    return np.array(value)
